# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/8

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The views for the unmatched offset management.

"""
from flask import Blueprint, render_template, redirect, url_for, flash

from accounting import db
from accounting.locale import lazy_gettext
from accounting.models import JournalEntryLineItem, Account
from accounting.utils.cast import s
from accounting.utils.offset_matcher import OffsetMatcher
from accounting.utils.pagination import Pagination
from accounting.utils.permission import has_permission, can_edit
from .queries import get_accounts_with_unmatched_offsets

bp: Blueprint = Blueprint("unmatched-offset", __name__)
"""The view blueprint for the unmatched offset management."""


@bp.get("", endpoint="dashboard")
@has_permission(can_edit)
def show_offset_dashboard() -> str:
    """Shows the dashboard about offsets.

    :return: The dashboard about offsets.
    """
    return render_template("accounting/unmatched-offset/dashboard.html",
                           list=get_accounts_with_unmatched_offsets())


@bp.get("<needOffsetAccount:account>", endpoint="list")
@has_permission(can_edit)
def show_unmatched_offsets(account: Account) -> str:
    """Shows the unmatched offsets in an account.

    :return: The unmatched offsets in an account.
    """
    matcher: OffsetMatcher = OffsetMatcher(account)
    pagination: Pagination \
        = Pagination[JournalEntryLineItem](matcher.unmatched_offsets,
                                           is_reversed=True)
    return render_template("accounting/unmatched-offset/list.html",
                           matcher=matcher,
                           list=pagination.list, pagination=pagination)


@bp.post("<needOffsetAccount:account>", endpoint="match")
@has_permission(can_edit)
def match_offsets(account: Account) -> redirect:
    """Matches the original line items with their offsets.

    :return: Redirection to the view of the unmatched offsets.
    """
    matcher: OffsetMatcher = OffsetMatcher(account)
    if not matcher.is_having_matches:
        flash(s(lazy_gettext("No more offset to match automatically.")),
              "success")
        return redirect(url_for("accounting.unmatched-offset.list",
                                account=account))
    matcher.match()
    db.session.commit()
    flash(s(lazy_gettext(
        "Matches %(matches)s from %(total)s unapplied line items.",
        matches=matcher.matches, total=matcher.total)), "success")
    return redirect(url_for("accounting.unmatched-offset.list",
                            account=account))



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/18

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The journal entry forms for the journal entry management.

"""
import datetime as dt
import typing as t
from abc import ABC, abstractmethod

import sqlalchemy as sa
from flask_babel import LazyString
from flask_wtf import FlaskForm
from wtforms import DateField, FieldList, FormField, TextAreaField, \
    BooleanField
from wtforms.validators import DataRequired, ValidationError

from accounting import db
from accounting.locale import lazy_gettext
from accounting.models import JournalEntry, Account, JournalEntryLineItem, \
    JournalEntryCurrency
from accounting.journal_entry.utils.account_option import AccountOption
from accounting.journal_entry.utils.original_line_items import \
    get_selectable_original_line_items
from accounting.journal_entry.utils.description_editor import DescriptionEditor
from accounting.utils.random_id import new_id
from accounting.utils.strip_text import strip_multiline_text
from accounting.utils.user import get_current_user_pk
from .currency import CurrencyForm, CashReceiptCurrencyForm, \
    CashDisbursementCurrencyForm, TransferCurrencyForm
from .line_item import LineItemForm, DebitLineItemForm, CreditLineItemForm
from .reorder import sort_journal_entries_in

DATE_REQUIRED: DataRequired = DataRequired(
    lazy_gettext("Please fill in the date."))
"""The validator to check if the date is empty."""


class NotBeforeOriginalLineItems:
    """The validator to check if the date is not before the
    original line items."""

    def __call__(self, form: FlaskForm, field: DateField) -> None:
        assert isinstance(form, JournalEntryForm)
        if field.data is None:
            return
        min_date: dt.date | None = form.min_date
        if min_date is None:
            return
        if field.data < min_date:
            raise ValidationError(lazy_gettext(
                "The date cannot be earlier than the original line items."))


class NotAfterOffsetItems:
    """The validator to check if the date is not after the offset items."""

    def __call__(self, form: FlaskForm, field: DateField) -> None:
        assert isinstance(form, JournalEntryForm)
        if field.data is None:
            return
        max_date: dt.date | None = form.max_date
        if max_date is None:
            return
        if field.data > max_date:
            raise ValidationError(lazy_gettext(
                "The date cannot be later than the offset items."))


class NeedSomeCurrencies:
    """The validator to check if there is any currency sub-form."""

    def __call__(self, form: FlaskForm, field: FieldList) -> None:
        if len(field) == 0:
            raise ValidationError(lazy_gettext("Please add some currencies."))


class CannotDeleteOriginalLineItemsWithOffset:
    """The validator to check the original line items with offset."""

    def __call__(self, form: FlaskForm, field: FieldList) -> None:
        assert isinstance(form, JournalEntryForm)
        if form.obj is None:
            return
        existing_matched_original_line_item_id: set[int] \
            = {x.id for x in form.obj.line_items if len(x.offsets) > 0}
        line_item_id_in_form: set[int] \
            = {x.id.data for x in form.line_items if x.id.data is not None}
        for line_item_id in existing_matched_original_line_item_id:
            if line_item_id not in line_item_id_in_form:
                raise ValidationError(lazy_gettext(
                    "Line items with offset cannot be deleted."))


class JournalEntryForm(FlaskForm):
    """The base form to create or edit a journal entry."""
    date = DateField()
    """The date."""
    currencies = FieldList(FormField(CurrencyForm))
    """The line items categorized by their currencies."""
    note = TextAreaField()
    """The note."""

    def __init__(self, *args, **kwargs):
        """Constructs a base journal entry form.

        :param args: The arguments.
        :param kwargs: The keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.is_modified: bool = False
        """Whether the journal entry is modified during populate_obj()."""
        self.collector: t.Type[LineItemCollector] = LineItemCollector
        """The line item collector.  The default is the base abstract
        collector only to provide the correct type.  The subclass forms should
        provide their own collectors."""
        self.obj: JournalEntry | None = kwargs.get("obj")
        """The journal entry, when editing an existing one."""
        self._is_need_payable: bool = False
        """Whether we need the payable original line items."""
        self._is_need_receivable: bool = False
        """Whether we need the receivable original line items."""
        self.__original_line_item_options: list[JournalEntryLineItem] | None \
            = None
        """The options of the original line items."""
        self.__net_balance_exceeded: dict[int, LazyString] | None = None
        """The original line items whose net balances were exceeded by the
        amounts in the line item sub-forms."""
        for line_item in self.line_items:
            line_item.journal_entry_form = self

    def populate_obj(self, obj: JournalEntry) -> None:
        """Populates the form data into a journal entry object.

        :param obj: The journal entry object.
        :return: None.
        """
        is_new: bool = obj.id is None
        if is_new:
            obj.id = new_id(JournalEntry)
        self.date: DateField
        self.__set_date(obj, self.date.data)
        obj.note = self.note.data

        collector_cls: t.Type[LineItemCollector] = self.collector
        collector: collector_cls = collector_cls(self, obj)
        collector.collect()

        to_delete: set[int] = {x.id for x in obj.line_items
                               if x.id not in collector.to_keep}
        if len(to_delete) > 0:
            JournalEntryLineItem.query\
                .filter(JournalEntryLineItem.id.in_(to_delete)).delete()
            self.is_modified = True

        if is_new or db.session.is_modified(obj):
            self.is_modified = True

        if is_new:
            current_user_pk: int = get_current_user_pk()
            obj.created_by_id = current_user_pk
            obj.updated_by_id = current_user_pk

    @property
    def line_items(self) -> list[LineItemForm]:
        """Collects and returns the line item sub-forms.

        :return: The line item sub-forms.
        """
        line_items: list[LineItemForm] = []
        for currency in self.currencies:
            line_items.extend(currency.line_items)
        return line_items

    def __set_date(self, obj: JournalEntry, new_date: dt.date) -> None:
        """Sets the journal entry date and number.

        :param obj: The journal entry object.
        :param new_date: The new date.
        :return: None.
        """
        if obj.date is None or obj.date != new_date:
            if obj.date is not None:
                sort_journal_entries_in(obj.date, obj.id)
            if self.max_date is not None and new_date == self.max_date:
                db_min_no: int | None = db.session.scalar(
                    sa.select(sa.func.min(JournalEntry.no))
                    .filter(JournalEntry.date == new_date))
                if db_min_no is None:
                    obj.date = new_date
                    obj.no = 1
                else:
                    obj.date = new_date
                    obj.no = db_min_no - 1
                    sort_journal_entries_in(new_date)
            else:
                sort_journal_entries_in(new_date, obj.id)
                count: int = JournalEntry.query\
                    .filter(JournalEntry.date == new_date).count()
                obj.date = new_date
                obj.no = count + 1

    @property
    def debit_account_options(self) -> list[AccountOption]:
        """The selectable debit accounts.

        :return: The selectable debit accounts.
        """
        accounts: list[AccountOption] \
            = [AccountOption(x) for x in Account.selectable_debit()
               if not (x.code[0] == "2" and x.is_need_offset)]
        in_use: set[int] = set(db.session.scalars(
            sa.select(JournalEntryLineItem.account_id)
            .filter(JournalEntryLineItem.is_debit)
            .group_by(JournalEntryLineItem.account_id)).all())
        for account in accounts:
            account.is_in_use = account.id in in_use
        return accounts

    @property
    def credit_account_options(self) -> list[AccountOption]:
        """The selectable credit accounts.

        :return: The selectable credit accounts.
        """
        accounts: list[AccountOption] \
            = [AccountOption(x) for x in Account.selectable_credit()
               if not (x.code[0] == "1" and x.is_need_offset)]
        in_use: set[int] = set(db.session.scalars(
            sa.select(JournalEntryLineItem.account_id)
            .filter(sa.not_(JournalEntryLineItem.is_debit))
            .group_by(JournalEntryLineItem.account_id)).all())
        for account in accounts:
            account.is_in_use = account.id in in_use
        return accounts

    @property
    def currencies_errors(self) -> list[str | LazyString]:
        """Returns the currency errors, without the errors in their sub-forms.

        :return: The currency errors, without the errors in their sub-forms.
        """
        return [x for x in self.currencies.errors
                if isinstance(x, str) or isinstance(x, LazyString)]

    @property
    def description_editor(self) -> DescriptionEditor:
        """Returns the description editor.

        :return: The description editor.
        """
        return DescriptionEditor()

    @property
    def original_line_item_options(self) -> list[JournalEntryLineItem]:
        """Returns the selectable original line items.

        :return: The selectable original line items.
        """
        if self.__original_line_item_options is None:
            self.__original_line_item_options \
                = get_selectable_original_line_items(
                    {x.id.data for x in self.line_items
                     if x.id.data is not None},
                    self._is_need_payable, self._is_need_receivable)
        return self.__original_line_item_options

    @property
    def min_date(self) -> dt.date | None:
        """Returns the minimal available date.

        :return: The minimal available date.
        """
        original_line_item_id: set[int] \
            = {x.original_line_item_id.data for x in self.line_items
               if x.original_line_item_id.data is not None}
        if len(original_line_item_id) == 0:
            return None
        select: sa.Select = sa.select(sa.func.max(JournalEntry.date))\
            .join(JournalEntryLineItem)\
            .filter(JournalEntryLineItem.id.in_(original_line_item_id))
        return db.session.scalar(select)

    @property
    def max_date(self) -> dt.date | None:
        """Returns the maximum available date.

        :return: The maximum available date.
        """
        line_item_id: set[int] = {x.id.data for x in self.line_items
                                  if x.id.data is not None}
        select: sa.Select = sa.select(sa.func.min(JournalEntry.date))\
            .join(JournalEntryLineItem)\
            .filter(JournalEntryLineItem.original_line_item_id
                    .in_(line_item_id))
        return db.session.scalar(select)


T = t.TypeVar("T", bound=JournalEntryForm)
"""A journal entry form variant."""


class LineItemCollector(t.Generic[T], ABC):
    """The line item collector."""

    def __init__(self, form: T, obj: JournalEntry):
        """Constructs the line item collector.

        :param form: The journal entry form.
        :param obj: The journal entry.
        """
        self.form: T = form
        """The journal entry form."""
        self.__obj: JournalEntry = obj
        """The journal entry object."""
        self.__line_items: list[JournalEntryLineItem] = list(obj.line_items)
        """The existing line items."""
        self.__line_items_by_id: dict[int, JournalEntryLineItem] \
            = {x.id: x for x in self.__line_items}
        """A dictionary from the line item ID to their line items."""
        self.__no_by_id: dict[int, int] \
            = {x.id: x.no for x in self.__line_items}
        """A dictionary from the line item number to their line items."""
        self.__currencies: list[JournalEntryCurrency] = obj.currencies
        """The currencies in the journal entry."""
        self._debit_no: int = 1
        """The number index for the debit line items."""
        self._credit_no: int = 1
        """The number index for the credit line items."""
        self.to_keep: set[int] = set()
        """The ID of the existing line items to keep."""

    @abstractmethod
    def collect(self) -> set[int]:
        """Collects the line items.

        :return: The ID of the line items to keep.
        """

    def _add_line_item(self, form: LineItemForm, currency_code: str, no: int) \
            -> None:
        """Composes a line item from the form.

        :param form: The line item form.
        :param currency_code: The code of the currency.
        :param no: The number of the line item.
        :return: None.
        """
        line_item: JournalEntryLineItem | None \
            = self.__line_items_by_id.get(form.id.data)
        if line_item is not None:
            line_item.currency_code = currency_code
            form.populate_obj(line_item)
            line_item.no = no
            if db.session.is_modified(line_item):
                self.form.is_modified = True
        else:
            line_item = JournalEntryLineItem()
            line_item.currency_code = currency_code
            form.populate_obj(line_item)
            line_item.no = no
            self.__obj.line_items.append(line_item)
            self.form.is_modified = True
        self.to_keep.add(line_item.id)

    def _make_cash_line_item(self, forms: list[LineItemForm], is_debit: bool,
                             currency_code: str, no: int) -> None:
        """Composes the cash line item at the other debit or credit of the
        cash journal entry.

        :param forms: The line item forms in the same currency.
        :param is_debit: True for a cash receipt journal entry, or False for a
            cash disbursement journal entry.
        :param currency_code: The code of the currency.
        :param no: The number of the line item.
        :return: None.
        """
        candidates: list[JournalEntryLineItem] \
            = [x for x in self.__line_items
               if x.is_debit == is_debit and x.currency_code == currency_code]
        line_item: JournalEntryLineItem
        if len(candidates) > 0:
            candidates.sort(key=lambda x: x.no)
            line_item = candidates[0]
            line_item.account_id = Account.cash().id
            line_item.description = None
            line_item.amount = sum([x.amount.data for x in forms])
            line_item.no = no
            if db.session.is_modified(line_item):
                self.form.is_modified = True
        else:
            line_item = JournalEntryLineItem()
            line_item.id = new_id(JournalEntryLineItem)
            line_item.is_debit = is_debit
            line_item.currency_code = currency_code
            line_item.account_id = Account.cash().id
            line_item.description = None
            line_item.amount = sum([x.amount.data for x in forms])
            line_item.no = no
            self.__obj.line_items.append(line_item)
            self.form.is_modified = True
        self.to_keep.add(line_item.id)

    def _sort_line_item_forms(self, forms: list[LineItemForm]) -> None:
        """Sorts the line item sub-forms.

        :param forms: The line item sub-forms.
        :return: None.
        """
        missing_no: int = 100 if len(self.__no_by_id) == 0 \
            else max(self.__no_by_id.values()) + 100
        ord_by_form: dict[LineItemForm, int] \
            = {forms[i]: i for i in range(len(forms))}
        recv_no: set[int] = {x.no.data for x in forms if x.no.data is not None}
        missing_recv_no: int = 100 if len(recv_no) == 0 else max(recv_no) + 100
        forms.sort(key=lambda x: (x.no.data or missing_recv_no,
                                  missing_no if x.id.data is None else
                                  self.__no_by_id.get(x.id.data, missing_no),
                                  ord_by_form.get(x)))

    def _sort_currency_forms(self, forms: list[CurrencyForm]) -> None:
        """Sorts the currency forms.

        :param forms: The currency forms.
        :return: None.
        """
        missing_no: int = len(self.__currencies) + 100
        no_by_code: dict[str, int] = {self.__currencies[i].code: i
                                      for i in range(len(self.__currencies))}
        ord_by_form: dict[CurrencyForm, int] \
            = {forms[i]: i for i in range(len(forms))}
        recv_no: set[int] = {x.no.data for x in forms if x.no.data is not None}
        missing_recv_no: int = 100 if len(recv_no) == 0 else max(recv_no) + 100
        forms.sort(key=lambda x: (x.no.data or missing_recv_no,
                                  no_by_code.get(x.code.data, missing_no),
                                  ord_by_form.get(x)))


class CashReceiptJournalEntryForm(JournalEntryForm):
    """The form to create or edit a cash receipt journal entry."""
    date = DateField(
        validators=[DATE_REQUIRED,
                    NotBeforeOriginalLineItems(),
                    NotAfterOffsetItems()])
    """The date."""
    currencies = FieldList(FormField(CashReceiptCurrencyForm), name="currency",
                           validators=[NeedSomeCurrencies()])
    """The line items categorized by their currencies."""
    note = TextAreaField(filters=[strip_multiline_text])
    """The note."""
    whole_form = BooleanField(
        validators=[CannotDeleteOriginalLineItemsWithOffset()])
    """The pseudo field for the whole form validators."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_need_receivable = True

        class Collector(LineItemCollector[CashReceiptJournalEntryForm]):
            """The line item collector for the cash receipt journal entries."""

            def collect(self) -> None:
                currencies: list[CashReceiptCurrencyForm] \
                    = [x.form for x in self.form.currencies]
                self._sort_currency_forms(currencies)
                for currency in currencies:
                    # The debit cash line item
                    self._make_cash_line_item(list(currency.credit), True,
                                              currency.code.data,
                                              self._debit_no)
                    self._debit_no = self._debit_no + 1

                    # The credit forms
                    credit_forms: list[CreditLineItemForm] \
                        = [x.form for x in currency.credit]
                    self._sort_line_item_forms(credit_forms)
                    for credit_form in credit_forms:
                        self._add_line_item(credit_form, currency.code.data,
                                            self._credit_no)
                        self._credit_no = self._credit_no + 1

        self.collector = Collector


class CashDisbursementJournalEntryForm(JournalEntryForm):
    """The form to create or edit a cash disbursement journal entry."""
    date = DateField(
        validators=[DATE_REQUIRED,
                    NotBeforeOriginalLineItems(),
                    NotAfterOffsetItems()])
    """The date."""
    currencies = FieldList(FormField(CashDisbursementCurrencyForm),
                           name="currency",
                           validators=[NeedSomeCurrencies()])
    """The line items categorized by their currencies."""
    note = TextAreaField(filters=[strip_multiline_text])
    """The note."""
    whole_form = BooleanField(
        validators=[CannotDeleteOriginalLineItemsWithOffset()])
    """The pseudo field for the whole form validators."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_need_payable = True

        class Collector(LineItemCollector[CashDisbursementJournalEntryForm]):
            """The line item collector for the cash disbursement journal
            entries."""

            def collect(self) -> None:
                currencies: list[CashDisbursementCurrencyForm] \
                    = [x.form for x in self.form.currencies]
                self._sort_currency_forms(currencies)
                for currency in currencies:
                    # The debit forms
                    debit_forms: list[DebitLineItemForm] \
                        = [x.form for x in currency.debit]
                    self._sort_line_item_forms(debit_forms)
                    for debit_form in debit_forms:
                        self._add_line_item(debit_form, currency.code.data,
                                            self._debit_no)
                        self._debit_no = self._debit_no + 1

                    # The credit forms
                    self._make_cash_line_item(list(currency.debit), False,
                                              currency.code.data,
                                              self._credit_no)
                    self._credit_no = self._credit_no + 1

        self.collector = Collector


class TransferJournalEntryForm(JournalEntryForm):
    """The form to create or edit a transfer journal entry."""
    date = DateField(
        validators=[DATE_REQUIRED,
                    NotBeforeOriginalLineItems(),
                    NotAfterOffsetItems()])
    """The date."""
    currencies = FieldList(FormField(TransferCurrencyForm), name="currency",
                           validators=[NeedSomeCurrencies()])
    """The line items categorized by their currencies."""
    note = TextAreaField(filters=[strip_multiline_text])
    """The note."""
    whole_form = BooleanField(
        validators=[CannotDeleteOriginalLineItemsWithOffset()])
    """The pseudo field for the whole form validators."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_need_payable = True
        self._is_need_receivable = True

        class Collector(LineItemCollector[TransferJournalEntryForm]):
            """The line item collector for the transfer journal entries."""

            def collect(self) -> None:
                currencies: list[TransferCurrencyForm] \
                    = [x.form for x in self.form.currencies]
                self._sort_currency_forms(currencies)
                for currency in currencies:
                    # The debit forms
                    debit_forms: list[DebitLineItemForm] \
                        = [x.form for x in currency.debit]
                    self._sort_line_item_forms(debit_forms)
                    for debit_form in debit_forms:
                        self._add_line_item(debit_form, currency.code.data,
                                            self._debit_no)
                        self._debit_no = self._debit_no + 1

                    # The credit forms
                    credit_forms: list[CreditLineItemForm] \
                        = [x.form for x in currency.credit]
                    self._sort_line_item_forms(credit_forms)
                    for credit_form in credit_forms:
                        self._add_line_item(credit_form, currency.code.data,
                                            self._credit_no)
                        self._credit_no = self._credit_no + 1

        self.collector = Collector



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/4

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The period specification parser.

"""
import calendar
import re
import typing as t
from datetime import date

from .period import Period
from .shortcuts import ThisMonth, LastMonth, SinceLastMonth, ThisYear, \
    LastYear, Today, Yesterday, AllTime

DATE_SPEC_RE: str = r"(\d{4})(?:-(\d{2})(?:-(\d{2}))?)?"
"""The regular expression of a date specification."""


def get_period(spec: str | None = None) -> Period:
    """Returns a period instance.

    :param spec: The period specification, or omit for the default.
    :return: The period instance.
    :raise ValueError: When the period specification is invalid.
    """
    if spec is None:
        return ThisMonth()
    named_periods: dict[str, t.Type[t.Callable[[], Period]]] = {
        "this-month": lambda: ThisMonth(),
        "last-month": lambda: LastMonth(),
        "since-last-month": lambda: SinceLastMonth(),
        "this-year": lambda: ThisYear(),
        "last-year": lambda: LastYear(),
        "today": lambda: Today(),
        "yesterday": lambda: Yesterday(),
        "all-time": lambda: AllTime(),
    }
    if spec in named_periods:
        return named_periods[spec]()
    start, end = __parse_spec(spec)
    if start is not None and end is not None and start > end:
        raise ValueError
    return Period(start, end)


def __parse_spec(text: str) -> tuple[date | None, date | None]:
    """Parses the period specification.

    :param text: The period specification.
    :return: The start and end day of the period.  The start and end day
        may be None.
    :raise ValueError: When the date is invalid.
    """
    if text == "-":
        return None, None
    m = re.match(f"^{DATE_SPEC_RE}$", text)
    if m is not None:
        return __get_start(m[1], m[2], m[3]), \
               __get_end(m[1], m[2], m[3])
    m = re.match(f"^{DATE_SPEC_RE}-$", text)
    if m is not None:
        return __get_start(m[1], m[2], m[3]), None
    m = re.match(f"-{DATE_SPEC_RE}$", text)
    if m is not None:
        return None, __get_end(m[1], m[2], m[3])
    m = re.match(f"^{DATE_SPEC_RE}-{DATE_SPEC_RE}$", text)
    if m is not None:
        return __get_start(m[1], m[2], m[3]), \
               __get_end(m[4], m[5], m[6])
    raise ValueError


def __get_start(year: str, month: str | None, day: str | None) -> date:
    """Returns the start of the period from the date representation.

    :param year: The year.
    :param month: The month, if any.
    :param day: The day, if any.
    :return: The start of the period.
    :raise ValueError: When the date is invalid.
    """
    if day is not None:
        return date(int(year), int(month), int(day))
    if month is not None:
        return date(int(year), int(month), 1)
    return date(int(year), 1, 1)


def __get_end(year: str, month: str | None, day: str | None) -> date:
    """Returns the end of the period from the date representation.

    :param year: The year.
    :param month: The month, if any.
    :param day: The day, if any.
    :return: The end of the period.
    :raise ValueError: When the date is invalid.
    """
    if day is not None:
        return date(int(year), int(month), int(day))
    if month is not None:
        year_n: int = int(year)
        month_n: int = int(month)
        day_n: int = calendar.monthrange(year_n, month_n)[1]
        return date(year_n, month_n, day_n)
    return date(int(year), 12, 31)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/25

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The accounting application.

"""
from pathlib import Path

from flask import Flask, Blueprint
from flask_sqlalchemy import SQLAlchemy

from accounting.utils.user import UserUtilityInterface

db: SQLAlchemy = SQLAlchemy()
"""The database instance."""
data_dir: Path = Path(__file__).parent / "data"
"""The data directory."""


def init_app(app: Flask, user_utils: UserUtilityInterface,
             url_prefix: str = "/accounting") -> None:
    """Initialize the application.

    :param app: The Flask application.
    :param user_utils: The user utilities.
    :param url_prefix: The URL prefix of the accounting application.
    :return: None.
    """
    # The database instance must be set before loading everything
    # in the application.
    global db
    db = app.extensions["sqlalchemy"]
    from .utils.user import init_user_utils
    init_user_utils(user_utils)

    bp: Blueprint = Blueprint("accounting", __name__,
                              template_folder="templates",
                              static_folder="static")

    from .template_filters import format_amount, format_date, default
    bp.add_app_template_filter(format_amount, "accounting_format_amount")
    bp.add_app_template_filter(format_date, "accounting_format_date")
    bp.add_app_template_filter(default, "accounting_default")

    from .template_globals import currency_options, default_currency_code
    bp.add_app_template_global(currency_options,
                               "accounting_currency_options")
    bp.add_app_template_global(default_currency_code,
                               "accounting_default_currency_code")

    from .commands import init_db_command
    app.cli.add_command(init_db_command)

    from . import locale
    locale.init_app(app, bp)

    from .utils import permission
    permission.init_app(bp, user_utils)

    from .utils import next_uri
    next_uri.init_app(bp)

    from . import base_account
    base_account.init_app(app, bp)

    from . import account
    account.init_app(app, bp)

    from . import currency
    currency.init_app(app, bp)

    from . import journal_entry
    journal_entry.init_app(app, bp)

    from . import report
    report.init_app(app, url_prefix)

    from . import option
    option.init_app(bp)

    from . import unmatched_offset
    unmatched_offset.init_app(bp)

    app.register_blueprint(bp, url_prefix=url_prefix)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/10

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The currency sub-forms for the journal entry management.

"""
from decimal import Decimal

import sqlalchemy as sa
from flask_babel import LazyString
from flask_wtf import FlaskForm
from wtforms import StringField, ValidationError, FieldList, IntegerField, \
    BooleanField, FormField
from wtforms.validators import DataRequired

from accounting import db
from accounting.forms import CurrencyExists
from accounting.locale import lazy_gettext
from accounting.models import JournalEntryLineItem
from accounting.utils.cast import be
from accounting.utils.offset_alias import offset_alias
from accounting.utils.strip_text import strip_text
from .line_item import LineItemForm, CreditLineItemForm, DebitLineItemForm

CURRENCY_REQUIRED: DataRequired = DataRequired(
    lazy_gettext("Please select the currency."))
"""The validator to check if the currency code is empty."""


class SameCurrencyAsOriginalLineItems:
    """The validator to check if the currency is the same as the
    original line items."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        assert isinstance(form, CurrencyForm)
        if field.data is None:
            return
        original_line_item_id: set[int] \
            = {x.original_line_item_id.data
               for x in form.line_items
               if x.original_line_item_id.data is not None}
        if len(original_line_item_id) == 0:
            return
        original_line_item_currency_codes: set[str] = set(db.session.scalars(
            sa.select(JournalEntryLineItem.currency_code)
            .filter(JournalEntryLineItem.id.in_(original_line_item_id))).all())
        for currency_code in original_line_item_currency_codes:
            if field.data != currency_code:
                raise ValidationError(lazy_gettext(
                    "The currency must be the same as the"
                    " original line item."))


class KeepCurrencyWhenHavingOffset:
    """The validator to check if the currency is the same when there is
    offset."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        assert isinstance(form, CurrencyForm)
        if field.data is None:
            return
        offset: sa.Alias = offset_alias()
        original_line_items: list[JournalEntryLineItem]\
            = JournalEntryLineItem.query\
            .join(offset, be(JournalEntryLineItem.id
                             == offset.c.original_line_item_id),
                  isouter=True)\
            .filter(JournalEntryLineItem.id
                    .in_({x.id.data for x in form.line_items
                          if x.id.data is not None}))\
            .group_by(JournalEntryLineItem.id,
                      JournalEntryLineItem.currency_code)\
            .having(sa.func.count(offset.c.id) > 0).all()
        for original_line_item in original_line_items:
            if original_line_item.currency_code != field.data:
                raise ValidationError(lazy_gettext(
                    "The currency must not be changed when there is offset."))


class NeedSomeLineItems:
    """The validator to check if there is any line item sub-form."""

    def __call__(self, form: FlaskForm, field: FieldList) -> None:
        if len(field) == 0:
            raise ValidationError(lazy_gettext(
                "Please add some line items."))


class IsBalanced:
    """The validator to check that the total amount of the debit and credit
    line items are equal."""

    def __call__(self, form: FlaskForm, field: BooleanField) -> None:
        assert isinstance(form, TransferCurrencyForm)
        if len(form.debit) == 0 or len(form.credit) == 0:
            return
        if form.debit_total != form.credit_total:
            raise ValidationError(lazy_gettext(
                "The totals of the debit and credit amounts do not match."))


class CurrencyForm(FlaskForm):
    """The form to create or edit a currency in a journal entry."""
    no = IntegerField()
    """The order in the journal entry."""
    code = StringField()
    """The currency code."""
    whole_form = BooleanField()
    """The pseudo field for the whole form validators."""

    @property
    def line_items(self) -> list[LineItemForm]:
        """Returns the line item sub-forms.

        :return: The line item sub-forms.
        """
        line_item_forms: list[LineItemForm] = []
        if isinstance(self, CashReceiptCurrencyForm):
            line_item_forms.extend([x.form for x in self.credit])
        elif isinstance(self, CashDisbursementCurrencyForm):
            line_item_forms.extend([x.form for x in self.debit])
        elif isinstance(self, TransferCurrencyForm):
            line_item_forms.extend([x.form for x in self.debit])
            line_item_forms.extend([x.form for x in self.credit])
        return line_item_forms

    @property
    def is_code_locked(self) -> bool:
        """Returns whether the currency code should not be changed.

        :return: True if the currency code should not be changed, or False
            otherwise
        """
        line_item_forms: list[LineItemForm] = self.line_items
        original_line_item_id: set[int] \
            = {x.original_line_item_id.data for x in line_item_forms
               if x.original_line_item_id.data is not None}
        if len(original_line_item_id) > 0:
            return True
        line_item_id: set[int] = {x.id.data for x in line_item_forms
                                  if x.id.data is not None}
        select: sa.Select = sa.select(sa.func.count(JournalEntryLineItem.id))\
            .filter(JournalEntryLineItem.original_line_item_id
                    .in_(line_item_id))
        return db.session.scalar(select) > 0


class CashReceiptCurrencyForm(CurrencyForm):
    """The form to create or edit a currency in a
    cash receipt journal entry."""
    no = IntegerField()
    """The order in the journal entry."""
    code = StringField(
        filters=[strip_text],
        validators=[CURRENCY_REQUIRED,
                    CurrencyExists(),
                    SameCurrencyAsOriginalLineItems(),
                    KeepCurrencyWhenHavingOffset()])
    """The currency code."""
    credit = FieldList(FormField(CreditLineItemForm),
                       validators=[NeedSomeLineItems()])
    """The credit line items."""
    whole_form = BooleanField()
    """The pseudo field for the whole form validators."""

    @property
    def credit_total(self) -> Decimal:
        """Returns the total amount of the credit line items.

        :return: The total amount of the credit line items.
        """
        return sum([x.amount.data for x in self.credit
                    if x.amount.data is not None])

    @property
    def credit_errors(self) -> list[str | LazyString]:
        """Returns the credit line item errors, without the errors in their
        sub-forms.

        :return:
        """
        return [x for x in self.credit.errors
                if isinstance(x, str) or isinstance(x, LazyString)]


class CashDisbursementCurrencyForm(CurrencyForm):
    """The form to create or edit a currency in a
    cash disbursement journal entry."""
    no = IntegerField()
    """The order in the journal entry."""
    code = StringField(
        filters=[strip_text],
        validators=[CURRENCY_REQUIRED,
                    CurrencyExists(),
                    SameCurrencyAsOriginalLineItems(),
                    KeepCurrencyWhenHavingOffset()])
    """The currency code."""
    debit = FieldList(FormField(DebitLineItemForm),
                      validators=[NeedSomeLineItems()])
    """The debit line items."""
    whole_form = BooleanField()
    """The pseudo field for the whole form validators."""

    @property
    def debit_total(self) -> Decimal:
        """Returns the total amount of the debit line items.

        :return: The total amount of the debit line items.
        """
        return sum([x.amount.data for x in self.debit
                    if x.amount.data is not None])

    @property
    def debit_errors(self) -> list[str | LazyString]:
        """Returns the debit line item errors, without the errors in their
        sub-forms.

        :return:
        """
        return [x for x in self.debit.errors
                if isinstance(x, str) or isinstance(x, LazyString)]


class TransferCurrencyForm(CurrencyForm):
    """The form to create or edit a currency in a transfer journal entry."""
    no = IntegerField()
    """The order in the journal entry."""
    code = StringField(
        filters=[strip_text],
        validators=[CURRENCY_REQUIRED,
                    CurrencyExists(),
                    SameCurrencyAsOriginalLineItems(),
                    KeepCurrencyWhenHavingOffset()])
    """The currency code."""
    debit = FieldList(FormField(DebitLineItemForm),
                      validators=[NeedSomeLineItems()])
    """The debit line items."""
    credit = FieldList(FormField(CreditLineItemForm),
                       validators=[NeedSomeLineItems()])
    """The credit line items."""
    whole_form = BooleanField(validators=[IsBalanced()])
    """The pseudo field for the whole form validators."""

    @property
    def debit_total(self) -> Decimal:
        """Returns the total amount of the debit line items.

        :return: The total amount of the debit line items.
        """
        return sum([x.amount.data for x in self.debit
                    if x.amount.data is not None])

    @property
    def credit_total(self) -> Decimal:
        """Returns the total amount of the credit line items.

        :return: The total amount of the credit line items.
        """
        return sum([x.amount.data for x in self.credit
                    if x.amount.data is not None])

    @property
    def debit_errors(self) -> list[str | LazyString]:
        """Returns the debit line item errors, without the errors in their
        sub-forms.

        :return:
        """
        return [x for x in self.debit.errors
                if isinstance(x, str) or isinstance(x, LazyString)]

    @property
    def credit_errors(self) -> list[str | LazyString]:
        """Returns the credit line item errors, without the errors in their
        sub-forms.

        :return:
        """
        return [x for x in self.credit.errors
                if isinstance(x, str) or isinstance(x, LazyString)]



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/30

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The queries for the account management.

"""
import sqlalchemy as sa
from flask import request

from accounting.locale import gettext
from accounting.models import Account, AccountL10n
from accounting.utils.query import parse_query_keywords


def get_account_query() -> list[Account]:
    """Returns the accounts, optionally filtered by the query.

    :return: The accounts.
    """
    keywords: list[str] = parse_query_keywords(request.args.get("q"))
    if len(keywords) == 0:
        return Account.query.order_by(Account.base_code, Account.no).all()
    code: sa.BinaryExpression = Account.base_code + "-" \
        + sa.func.substr("000" + sa.cast(Account.no, sa.String),
                         sa.func.char_length(sa.cast(Account.no,
                                                     sa.String)) + 1)
    conditions: list[sa.BinaryExpression] = []
    for k in keywords:
        l10n: list[AccountL10n] = AccountL10n.query\
            .filter(AccountL10n.title.icontains(k)).all()
        l10n_matches: set[str] = {x.account_id for x in l10n}
        sub_conditions: list[sa.BinaryExpression] \
            = [Account.base_code.contains(k),
               Account.title_l10n.icontains(k),
               code.contains(k),
               Account.id.in_(l10n_matches)]
        if k in gettext("Needs Offset"):
            sub_conditions.append(Account.is_need_offset)
        conditions.append(sa.or_(*sub_conditions))

    return Account.query.filter(*conditions)\
        .order_by(Account.base_code, Account.no).all()



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/6

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The views for the currency management.

"""
from urllib.parse import urlencode, parse_qsl

import sqlalchemy as sa
from flask import Blueprint, render_template, redirect, session, request, \
    flash, url_for
from werkzeug.datastructures import ImmutableMultiDict

from accounting import db
from accounting.locale import lazy_gettext
from accounting.models import Currency
from accounting.utils.cast import s
from accounting.utils.flash_errors import flash_form_errors
from accounting.utils.next_uri import inherit_next, or_next
from accounting.utils.pagination import Pagination
from accounting.utils.permission import has_permission, can_view, can_edit
from accounting.utils.user import get_current_user_pk
from .forms import CurrencyForm
from .queries import get_currency_query

bp: Blueprint = Blueprint("currency", __name__)
"""The view blueprint for the currency management."""
api_bp: Blueprint = Blueprint("currency-api", __name__)
"""The view blueprint for the currency management API."""


@bp.get("", endpoint="list")
@has_permission(can_view)
def list_currencies() -> str:
    """Lists the currencies.

    :return: The currency list.
    """
    currencies: list[Currency] = get_currency_query()
    pagination: Pagination = Pagination[Currency](currencies)
    return render_template("accounting/currency/list.html",
                           list=pagination.list, pagination=pagination)


@bp.get("create", endpoint="create")
@has_permission(can_edit)
def show_add_currency_form() -> str:
    """Shows the form to add a currency.

    :return: The form to add a currency.
    """
    if "form" in session:
        form = CurrencyForm(ImmutableMultiDict(parse_qsl(session["form"])))
        del session["form"]
        form.validate()
    else:
        form = CurrencyForm()
    return render_template("accounting/currency/create.html",
                           form=form)


@bp.post("store", endpoint="store")
@has_permission(can_edit)
def add_currency() -> redirect:
    """Adds a currency.

    :return: The redirection to the currency detail on success, or the currency
        creation form on error.
    """
    form = CurrencyForm(request.form)
    if not form.validate():
        flash_form_errors(form)
        session["form"] = urlencode(list(request.form.items()))
        return redirect(inherit_next(url_for("accounting.currency.create")))
    currency: Currency = Currency()
    form.populate_obj(currency)
    db.session.add(currency)
    db.session.commit()
    flash(s(lazy_gettext("The currency is added successfully.")), "success")
    return redirect(inherit_next(__get_detail_uri(currency)))


@bp.get("<currency:currency>", endpoint="detail")
@has_permission(can_view)
def show_currency_detail(currency: Currency) -> str:
    """Shows the currency detail.

    :param currency: The currency.
    :return: The detail.
    """
    return render_template("accounting/currency/detail.html", obj=currency)


@bp.get("<currency:currency>/edit", endpoint="edit")
@has_permission(can_edit)
def show_currency_edit_form(currency: Currency) -> str:
    """Shows the form to edit a currency.

    :param currency: The currency.
    :return: The form to edit the currency.
    """
    form: CurrencyForm
    if "form" in session:
        form = CurrencyForm(ImmutableMultiDict(parse_qsl(session["form"])))
        del session["form"]
        form.validate()
    else:
        form = CurrencyForm(obj=currency)
    return render_template("accounting/currency/edit.html",
                           currency=currency, form=form)


@bp.post("<currency:currency>/update", endpoint="update")
@has_permission(can_edit)
def update_currency(currency: Currency) -> redirect:
    """Updates a currency.

    :param currency: The currency.
    :return: The redirection to the currency detail on success, or the currency
        edit form on error.
    """
    form = CurrencyForm(request.form)
    form.obj_code = currency.code
    if not form.validate():
        flash_form_errors(form)
        session["form"] = urlencode(list(request.form.items()))
        return redirect(inherit_next(url_for("accounting.currency.edit",
                                             currency=currency)))
    with db.session.no_autoflush:
        form.populate_obj(currency)
    if not currency.is_modified:
        flash(s(lazy_gettext("The currency was not modified.")), "success")
        return redirect(inherit_next(__get_detail_uri(currency)))
    currency.updated_by_id = get_current_user_pk()
    currency.updated_at = sa.func.now()
    db.session.commit()
    flash(s(lazy_gettext("The currency is updated successfully.")), "success")
    return redirect(inherit_next(__get_detail_uri(currency)))


@bp.post("<currency:currency>/delete", endpoint="delete")
@has_permission(can_edit)
def delete_currency(currency: Currency) -> redirect:
    """Deletes a currency.

    :param currency: The currency.
    :return: The redirection to the currency list on success, or the currency
        detail on error.
    """
    if not currency.can_delete:
        flash(s(lazy_gettext("The currency cannot be deleted.")), "error")
        return redirect(inherit_next(__get_detail_uri(currency)))
    currency.delete()
    db.session.commit()
    flash(s(lazy_gettext("The currency is deleted successfully.")), "success")
    return redirect(or_next(url_for("accounting.currency.list")))


@api_bp.get("exists-code", endpoint="exists")
@has_permission(can_edit)
def exists_code() -> dict[str, bool]:
    """Validates whether a currency code exists.

    :return: Whether the currency code exists.
    """
    return {"exists": db.session.get(Currency, request.args["q"]) is not None}


def __get_detail_uri(currency: Currency) -> str:
    """Returns the detail URI of a currency.

    :param currency: The currency.
    :return: The detail URI of the currency.
    """
    return url_for("accounting.currency.detail", currency=currency)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/1

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The utilities to handle the next URI.

This module should not import any other module from the application.

"""
from urllib.parse import urlparse, parse_qsl, ParseResult, urlencode, \
    urlunparse

from flask import request, Blueprint


def append_next(uri: str) -> str:
    """Appends the current URI as the next URI to the query argument.

    :param uri: The URI.
    :return: The URI with the current URI appended as the next URI.
    """
    next_uri: str = request.full_path if request.query_string else request.path
    return __set_next(uri, next_uri)


def inherit_next(uri: str) -> str:
    """Inherits the current next URI to the query argument, if exists.

    :param uri: The URI.
    :return: The URI with the current next URI added at the query argument.
    """
    next_uri: str | None = request.form.get("next") \
        if request.method == "POST" else request.args.get("next")
    if next_uri is None:
        return uri
    return __set_next(uri, next_uri)


def or_next(uri: str) -> str:
    """Returns the next URI, if exists, or the supplied URI.

    :param uri: The URI.
    :return: The next URI or the supplied URI.
    """
    next_uri: str | None = request.form.get("next") \
        if request.method == "POST" else request.args.get("next")
    return uri if next_uri is None else next_uri


def __set_next(uri: str, next_uri: str) -> str:
    """Sets the next URI to the query arguments.

    :param uri: The URI.
    :param next_uri: The next URI.
    :return: The URI with the next URI set.
    """
    uri_p: ParseResult = urlparse(uri)
    params: list[tuple[str, str]] = parse_qsl(uri_p.query)
    params = [x for x in params if x[0] != "next"]
    params.append(("next", next_uri))
    parts: list[str] = list(uri_p)
    parts[4] = urlencode(params)
    return urlunparse(parts)


def init_app(bp: Blueprint) -> None:
    """Initializes the application.

    :param bp: The blueprint of the accounting application.
    :return: None.
    """
    bp.add_app_template_filter(append_next, "accounting_append_next")
    bp.add_app_template_filter(inherit_next, "accounting_inherit_next")
    bp.add_app_template_filter(or_next, "accounting_or_next")



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/10

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The test for the console commands.

"""
import csv
import typing as t
import unittest

import sqlalchemy as sa
from click.testing import Result
from flask import Flask
from flask.testing import FlaskCliRunner
from sqlalchemy.sql.ddl import DropTable

from test_site import db
from testlib import create_test_app


class ConsoleCommandTestCase(unittest.TestCase):
    """The console command test case."""

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        self.app: Flask = create_test_app()

        with self.app.app_context():
            # Drop every accounting table, to see if accounting-init recreates
            # them correctly.
            tables: list[sa.Table] \
                = [db.metadata.tables[x] for x in db.metadata.tables
                   if x.startswith("accounting_")]
            for table in tables:
                db.session.execute(DropTable(table))
            db.session.commit()
            inspector: sa.Inspector = sa.inspect(db.session.connection())
            self.assertEqual(len({x for x in inspector.get_table_names()
                                  if x.startswith("accounting_")}),
                             0)

    def test_init(self) -> None:
        """Tests the "accounting-init" console command.

        :return: None.
        """
        runner: FlaskCliRunner = self.app.test_cli_runner()
        with self.app.app_context():
            result: Result = runner.invoke(
                args=["accounting-init-db", "-u", "editor"])
        self.assertEqual(result.exit_code, 0,
                         result.output + str(result.exception))
        self.__test_base_account_data()
        self.__test_account_data()
        self.__test_currency_data()

    def __test_base_account_data(self) -> None:
        """Tests the base account data.

        :return: None.
        """
        from accounting import data_dir
        from accounting.models import BaseAccount

        with open(data_dir / "base_accounts.csv") as fp:
            data: dict[dict[str, t.Any]] \
                = {x["code"]: {"code": x["code"],
                               "title": x["title"],
                               "l10n": {y[5:]: x[y]
                                        for y in x if y.startswith("l10n-")}}
                   for x in csv.DictReader(fp)}

        with self.app.app_context():
            accounts: list[BaseAccount] = BaseAccount.query.all()

        self.assertEqual(len(accounts), len(data))
        for account in accounts:
            self.assertIn(account.code, data)
            self.assertEqual(account.title_l10n, data[account.code]["title"])
            l10n: dict[str, str] = {x.locale: x.title for x in account.l10n}
            self.assertEqual(len(l10n), len(data[account.code]["l10n"]))
            for locale in l10n:
                self.assertIn(locale, data[account.code]["l10n"])
                self.assertEqual(l10n[locale],
                                 data[account.code]["l10n"][locale])

    def __test_account_data(self) -> None:
        """Tests the account data.

        :return: None.
        """
        from accounting.models import BaseAccount, Account, AccountL10n

        with self.app.app_context():
            bases: list[BaseAccount] = BaseAccount.query\
                .filter(sa.func.char_length(BaseAccount.code) == 4).all()
            accounts: list[Account] = Account.query.all()
            l10n: list[AccountL10n] = AccountL10n.query.all()

        self.assertEqual({x.code for x in bases},
                         {x.base_code for x in accounts})
        self.assertEqual(len(accounts), len(bases))
        self.assertEqual(len(l10n), len(bases) * 2)
        base_dict: dict[str, BaseAccount] = {x.code: x for x in bases}
        for account in accounts:
            base: BaseAccount = base_dict[account.base_code]
            self.assertEqual(account.no, 1)
            self.assertEqual(account.title_l10n, base.title_l10n)
            self.assertEqual({x.locale: x.title for x in account.l10n},
                             {x.locale: x.title for x in base.l10n})

    def __test_currency_data(self) -> None:
        """Tests the currency data.

        :return: None.
        """
        from accounting import data_dir
        from accounting.models import Currency

        with open(data_dir / "currencies.csv") as fp:
            data: dict[dict[str, t.Any]] \
                = {x["code"]: {"code": x["code"],
                               "name": x["name"],
                               "l10n": {y[5:]: x[y]
                                        for y in x if y.startswith("l10n-")}}
                   for x in csv.DictReader(fp)}

        with self.app.app_context():
            currencies: list[Currency] = Currency.query.all()

        self.assertEqual(len(currencies), len(data))
        for currency in currencies:
            self.assertIn(currency.code, data)
            self.assertEqual(currency.name_l10n, data[currency.code]["name"])
            l10n: dict[str, str] = {x.locale: x.name for x in currency.l10n}
            self.assertEqual(len(l10n), len(data[currency.code]["l10n"]))
            for locale in l10n:
                self.assertIn(locale, data[currency.code]["l10n"])
                self.assertEqual(l10n[locale],
                                 data[currency.code]["l10n"][locale])



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/3

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The path converters for the report management.

"""
import re

from flask import abort
from werkzeug.routing import BaseConverter

from accounting.models import Account
from accounting.utils.current_account import CurrentAccount
from .period import Period, get_period


class PeriodConverter(BaseConverter):
    """The converter to convert the period specification from and to the
    corresponding period in the routes."""

    def to_python(self, value: str) -> Period:
        """Converts a period specification to a period.

        :param value: The period specification.
        :return: The corresponding period.
        """
        try:
            return get_period(value)
        except ValueError:
            abort(404)

    def to_url(self, value: Period) -> str:
        """Converts a period to its specification.

        :param value: The period.
        :return: Its specification.
        """
        return value.spec


class CurrentAccountConverter(BaseConverter):
    """The converter to convert the current account code from and to the
    corresponding account in the routes."""

    def to_python(self, value: str) -> CurrentAccount:
        """Converts an account code to an account.

        :param value: The account code.
        :return: The corresponding account.
        """
        if value == CurrentAccount.CURRENT_AL_CODE:
            return CurrentAccount.current_assets_and_liabilities()
        if not re.match("^[12][12]", value):
            abort(404)
        account: Account | None = Account.find_by_code(value)
        if account is None:
            abort(404)
        return CurrentAccount(account)

    def to_url(self, value: CurrentAccount) -> str:
        """Converts an account to account code.

        :param value: The account.
        :return: Its code.
        """
        return value.code


class NeedOffsetAccountConverter(BaseConverter):
    """The converter to convert the unapplied original line item account code
    from and to the corresponding account in the routes."""

    def to_python(self, value: str) -> Account:
        """Converts an account code to an account.

        :param value: The account code.
        :return: The corresponding account.
        """
        account: Account | None = Account.find_by_code(value)
        if account is None:
            abort(404)
        if not account.is_need_offset:
            abort(404)
        return account

    def to_url(self, value: Account) -> str:
        """Converts an account to account code.

        :param value: The account.
        :return: Its code.
        """
        return value.code



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/6

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The currency management.

"""
from flask import Flask, Blueprint

from .commands import init_currencies_command


def init_app(app: Flask, bp: Blueprint) -> None:
    """Initialize the application.

    :param app: The Flask application.
    :param bp: The blueprint of the accounting application.
    :return: None.
    """
    from .converters import CurrencyConverter
    app.url_map.converters["currency"] = CurrencyConverter

    from .views import bp as currency_bp, api_bp as currency_api_bp
    bp.register_blueprint(currency_bp, url_prefix="/currencies")
    bp.register_blueprint(currency_api_bp, url_prefix="/api/currencies")



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/25

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The permissions.

This module should not import any other module from the application.

"""
import typing as t

from flask import abort, Blueprint, Response

from accounting.utils.user import get_current_user, UserUtilityInterface


def has_permission(rule: t.Callable[[], bool]) -> t.Callable:
    """The permission decorator to check whether the current user is allowed.

    :param rule: The permission rule.
    :return: The view decorator.
    """

    def decorator(view: t.Callable) -> t.Callable:
        """The view decorator to decorate a view with permission tests.

        :param view: The view.
        :return: The decorated view.
        """

        def decorated_view(*args, **kwargs):
            """The decorated view that tests against a permission rule.

            :param args: The arguments of the view.
            :param kwargs: The keyword arguments of the view.
            :return: The response of the view.
            :raise Forbidden: When the user is denied.
            """
            if not rule():
                if get_current_user() is None:
                    response: Response | None = _unauthorized_func()
                    if response is not None:
                        return response
                abort(403)
            return view(*args, **kwargs)

        return decorated_view

    return decorator


__can_view_func: t.Callable[[], bool] = lambda: True
"""The callback that returns whether the current user can view the accounting
data."""
__can_edit_func: t.Callable[[], bool] = lambda: True
"""The callback that returns whether the current user can edit the accounting
data."""
__can_admin_func: t.Callable[[], bool] = lambda: True
"""The callback that returns whether the current user can administrate the
accounting settings."""
_unauthorized_func: t.Callable[[], Response | None] \
    = lambda: Response(status=403)
"""The callback that returns the response to require the user to log in."""


def can_view() -> bool:
    """Returns whether the current user can view the accounting data.

    :return: True if the current user can view the accounting data, or False
        otherwise.
    """
    return __can_view_func()


def can_edit() -> bool:
    """Returns whether the current user can edit the accounting data.

    The user has to log in.

    :return: True if the current user can edit the accounting data, or False
        otherwise.
    """
    if get_current_user() is None:
        return False
    return __can_edit_func()


def can_admin() -> bool:
    """Returns whether the current user can administrate the accounting
    settings.

    The user has to log in.

    :return: True if the current user can administrate the accounting settings,
        or False otherwise.
    """
    if get_current_user() is None:
        return False
    return __can_admin_func()


def init_app(bp: Blueprint, user_utils: UserUtilityInterface) -> None:
    """Initializes the application.

    :param bp: The blueprint of the accounting application.
    :param user_utils: The user utilities.
    :return: None.
    """
    global __can_view_func, __can_edit_func, __can_admin_func, \
        _unauthorized_func
    __can_view_func = user_utils.can_view
    __can_edit_func = user_utils.can_edit
    __can_admin_func = user_utils.can_admin
    _unauthorized_func = user_utils.unauthorized
    bp.add_app_template_global(user_utils.can_view, "accounting_can_view")
    bp.add_app_template_global(user_utils.can_edit, "accounting_can_edit")
    bp.add_app_template_global(user_utils.can_admin, "accounting_can_admin")



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/7

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The income statement.

"""
from decimal import Decimal

import sqlalchemy as sa
from flask import render_template, Response

from accounting import db
from accounting.locale import gettext
from accounting.models import Currency, BaseAccount, Account, JournalEntry, \
    JournalEntryLineItem
from accounting.report.period import Period, PeriodChooser
from accounting.report.utils.base_page_params import BasePageParams
from accounting.report.utils.base_report import BaseReport
from accounting.report.utils.csv_export import BaseCSVRow, csv_download, \
    period_spec
from accounting.report.utils.option_link import OptionLink
from accounting.report.utils.report_chooser import ReportChooser
from accounting.report.utils.report_type import ReportType
from accounting.report.utils.urls import ledger_url, income_statement_url


class ReportAccount:
    """An account in the report."""

    def __init__(self, account: Account, amount: Decimal, url: str):
        """Constructs an account in the report.

        :param account: The account.
        :param amount: The amount.
        :param url: The URL to the ledger of the account.
        """
        self.account: Account = account
        """The account."""
        self.amount: Decimal = amount
        """The amount of the account."""
        self.url: str = url
        """The URL to the ledger of the account."""


class AccumulatedTotal:
    """An accumulated total."""

    def __init__(self, title: str):
        """Constructs an accumulated total.

        :param title: The title.
        """
        self.title: str = title
        """The account."""
        self.amount: Decimal = Decimal("0")
        """The amount of the account."""


class Subsection:
    """A subsection."""

    def __init__(self, title: BaseAccount):
        """Constructs a subsection.

        :param title: The title account.
        """
        self.title: BaseAccount = title
        """The title account."""
        self.accounts: list[ReportAccount] = []
        """The accounts in the subsection."""

    @property
    def total(self) -> Decimal:
        """Returns the total of the subsection.

        :return: The total of the subsection.
        """
        return sum([x.amount for x in self.accounts])


class Section:
    """A section."""

    def __init__(self, title: BaseAccount, accumulated_title: str):
        """Constructs a section.

        :param title: The title account.
        :param accumulated_title: The title for the accumulated total.
        """
        self.title: BaseAccount = title
        """The title account."""
        self.subsections: list[Subsection] = []
        """The subsections in the section."""
        self.accumulated: AccumulatedTotal \
            = AccumulatedTotal(accumulated_title)

    @property
    def total(self) -> Decimal:
        """Returns the total of the section.

        :return: The total of the section.
        """
        return sum([x.total for x in self.subsections])


class CSVRow(BaseCSVRow):
    """A row in the CSV."""

    def __init__(self, text: str | None, amount: str | Decimal | None):
        """Constructs a row in the CSV.

        :param text: The text.
        :param amount: The amount.
        """
        self.text: str | None = text
        """The text."""
        self.amount: str | Decimal | None = amount
        """The amount."""

    @property
    def values(self) -> list[str | Decimal | None]:
        """Returns the values of the row.

        :return: The values of the row.
        """
        return [self.text, self.amount]


class PageParams(BasePageParams):
    """The HTML page parameters."""

    def __init__(self, currency: Currency,
                 period: Period,
                 has_data: bool,
                 sections: list[Section], ):
        """Constructs the HTML page parameters.

        :param currency: The currency.
        :param period: The period.
        :param has_data: True if there is any data, or False otherwise.
        """
        self.currency: Currency = currency
        """The currency."""
        self.period: Period = period
        """The period."""
        self.__has_data: bool = has_data
        """True if there is any data, or False otherwise."""
        self.sections: list[Section] = sections
        """The sections in the income statement."""
        self.period_chooser: PeriodChooser = PeriodChooser(
            lambda x: income_statement_url(currency, x))
        """The period chooser."""

    @property
    def has_data(self) -> bool:
        """Returns whether there is any data on the page.

        :return: True if there is any data, or False otherwise.
        """
        return self.__has_data

    @property
    def report_chooser(self) -> ReportChooser:
        """Returns the report chooser.

        :return: The report chooser.
        """
        return ReportChooser(ReportType.INCOME_STATEMENT,
                             currency=self.currency,
                             period=self.period)

    @property
    def currency_options(self) -> list[OptionLink]:
        """Returns the currency options.

        :return: The currency options.
        """
        return self._get_currency_options(
            lambda x: income_statement_url(x, self.period), self.currency)


class IncomeStatement(BaseReport):
    """The income statement."""

    def __init__(self, currency: Currency, period: Period):
        """Constructs an income statement.

        :param currency: The currency.
        :param period: The period.
        """
        self.__currency: Currency = currency
        """The currency."""
        self.__period: Period = period
        """The period."""
        self.__has_data: bool
        """True if there is any data, or False otherwise."""
        self.__sections: list[Section]
        """The sections."""
        self.__set_data()

    def __set_data(self) -> None:
        """Queries and sets data sections in the income statement.

        :return: None.
        """
        balances: list[ReportAccount] = self.__query_balances()

        titles: list[BaseAccount] = BaseAccount.query\
            .filter(BaseAccount.code.in_({"4", "5", "6", "7", "8", "9"})).all()
        subtitles: list[BaseAccount] = BaseAccount.query\
            .filter(BaseAccount.code.in_({x.account.base_code[:2]
                                          for x in balances})).all()

        total_titles: dict[str, str] \
            = {"4": gettext("total operating revenue"),
               "5": gettext("gross income"),
               "6": gettext("operating income"),
               "7": gettext("before tax income"),
               "8": gettext("after tax income"),
               "9": gettext("net income or loss for current period")}

        sections: dict[str, Section] \
            = {x.code: Section(x, total_titles[x.code]) for x in titles}
        subsections: dict[str, Subsection] \
            = {x.code: Subsection(x) for x in subtitles}
        for subsection in subsections.values():
            sections[subsection.title.code[0]].subsections.append(subsection)
        for balance in balances:
            subsections[balance.account.base_code[:2]].accounts.append(balance)

        self.__has_data = len(balances) > 0
        self.__sections = sorted(sections.values(), key=lambda x: x.title.code)
        total: Decimal = Decimal("0")
        for section in self.__sections:
            total = total + section.total
            section.accumulated.amount = total

    def __query_balances(self) -> list[ReportAccount]:
        """Queries and returns the balances.

        :return: The balances.
        """
        sub_conditions: list[sa.BinaryExpression] \
            = [Account.base_code.startswith(str(x)) for x in range(4, 10)]
        conditions: list[sa.BinaryExpression] \
            = [JournalEntryLineItem.currency_code == self.__currency.code,
               sa.or_(*sub_conditions)]
        if self.__period.start is not None:
            conditions.append(JournalEntry.date >= self.__period.start)
        if self.__period.end is not None:
            conditions.append(JournalEntry.date <= self.__period.end)
        balance_func: sa.Function = sa.func.sum(sa.case(
            (JournalEntryLineItem.is_debit, -JournalEntryLineItem.amount),
            else_=JournalEntryLineItem.amount)).label("balance")
        select_balances: sa.Select = sa.select(Account.id, balance_func)\
            .join(JournalEntry).join(Account)\
            .filter(*conditions)\
            .group_by(Account.id)\
            .having(balance_func != 0)\
            .order_by(Account.base_code, Account.no)
        balances: list[sa.Row] = db.session.execute(select_balances).all()
        accounts: dict[int, Account] \
            = {x.id: x for x in Account.query
               .filter(Account.id.in_([x.id for x in balances])).all()}
        return [ReportAccount(account=accounts[x.id],
                              amount=x.balance,
                              url=ledger_url(self.__currency,
                                             accounts[x.id],
                                             self.__period))
                for x in balances]

    def csv(self) -> Response:
        """Returns the report as CSV for download.

        :return: The response of the report for download.
        """
        filename: str = "income-statement-{currency}-{period}.csv"\
            .format(currency=self.__currency.code,
                    period=period_spec(self.__period))
        return csv_download(filename, self.__get_csv_rows())

    def __get_csv_rows(self) -> list[CSVRow]:
        """Composes and returns the CSV rows.

        :return: The CSV rows.
        """
        total_str: str = gettext("Total")
        rows: list[CSVRow] = [CSVRow(None, gettext("Amount"))]
        for section in self.__sections:
            rows.append(CSVRow(str(section.title).title(), None))
            for subsection in section.subsections:
                rows.append(CSVRow(f" {str(subsection.title).title()}", None))
                for account in subsection.accounts:
                    rows.append(CSVRow(f"  {str(account.account).title()}",
                                       account.amount))
                rows.append(CSVRow(f" {total_str}", subsection.total))
            rows.append(CSVRow(section.accumulated.title.title(),
                               section.accumulated.amount))
            rows.append(CSVRow(None, None))
        rows = rows[:-1]
        return rows

    def html(self) -> str:
        """Composes and returns the report as HTML.

        :return: The report as HTML.
        """
        params: PageParams = PageParams(currency=self.__currency,
                                        period=self.__period,
                                        has_data=self.__has_data,
                                        sections=self.__sections)
        return render_template("accounting/report/income-statement.html",
                               report=params)



# The Mia! Accounting Demonstration Website.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/13

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The common library for the Mia! Accounting demonstration website.

"""
from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod
from datetime import date, timedelta
from decimal import Decimal
from secrets import randbelow

import sqlalchemy as sa
from flask import Flask

from . import db
from .auth import User


class Accounts:
    """The shortcuts to the common accounts."""
    CASH: str = "1111-001"
    BANK: str = "1113-001"
    RECEIVABLE: str = "1141-001"
    MACHINERY: str = "1441-001"
    PAYABLE: str = "2141-001"
    SERVICE: str = "4611-001"
    RENT_EXPENSE: str = "6252-001"
    MEAL: str = "6272-001"


class JournalEntryLineItemData:
    """The journal entry line item data."""

    def __init__(self, account: str, description: str | None, amount: str,
                 original_line_item: JournalEntryLineItemData | None = None):
        """Constructs the journal entry line item data.

        :param account: The account code.
        :param description: The description.
        :param amount: The amount.
        :param original_line_item: The original journal entry line item.
        """
        self.journal_entry: JournalEntryData | None = None
        self.id: int = -1
        self.no: int = -1
        self.original_line_item: JournalEntryLineItemData | None \
            = original_line_item
        self.account: str = account
        self.description: str | None = description
        self.amount: Decimal = Decimal(amount)

    def form(self, prefix: str, debit_credit: str, index: int,
             is_update: bool) -> dict[str, str]:
        """Returns the line item as form data.

        :param prefix: The prefix of the form fields.
        :param debit_credit: Either "debit" or "credit".
        :param index: The line item index.
        :param is_update: True for an update operation, or False otherwise
        :return: The form data.
        """
        prefix = f"{prefix}-{debit_credit}-{index}"
        form: dict[str, str] = {f"{prefix}-account_code": self.account,
                                f"{prefix}-description": self.description,
                                f"{prefix}-amount": str(self.amount)}
        if is_update and self.id != -1:
            form[f"{prefix}-id"] = str(self.id)
        form[f"{prefix}-no"] = str(index) if self.no == -1 else str(self.no)
        if self.original_line_item is not None:
            assert self.original_line_item.id != -1
            form[f"{prefix}-original_line_item_id"] \
                = str(self.original_line_item.id)
        return form


class JournalEntryCurrencyData:
    """The journal entry currency data."""

    def __init__(self, currency: str, debit: list[JournalEntryLineItemData],
                 credit: list[JournalEntryLineItemData]):
        """Constructs the journal entry currency data.

        :param currency: The currency code.
        :param debit: The debit line items.
        :param credit: The credit line items.
        """
        self.code: str = currency
        self.debit: list[JournalEntryLineItemData] = debit
        self.credit: list[JournalEntryLineItemData] = credit

    def form(self, index: int, is_update: bool) -> dict[str, str]:
        """Returns the currency as form data.

        :param index: The currency index.
        :param is_update: True for an update operation, or False otherwise
        :return: The form data.
        """
        prefix: str = f"currency-{index}"
        form: dict[str, str] = {f"{prefix}-code": self.code}
        for i in range(len(self.debit)):
            form.update(self.debit[i].form(prefix, "debit", i + 1, is_update))
        for i in range(len(self.credit)):
            form.update(self.credit[i].form(prefix, "credit", i + 1,
                                            is_update))
        return form


class JournalEntryData:
    """The journal entry data."""

    def __init__(self, days: int, currencies: list[JournalEntryCurrencyData]):
        """Constructs a journal entry.

        :param days: The number of days before today.
        :param currencies: The journal entry currency data.
        """
        self.id: int = -1
        self.days: int = days
        self.currencies: list[JournalEntryCurrencyData] = currencies
        self.note: str | None = None
        for currency in self.currencies:
            for line_item in currency.debit:
                line_item.journal_entry = self
            for line_item in currency.credit:
                line_item.journal_entry = self

    def new_form(self, csrf_token: str, next_uri: str) -> dict[str, str]:
        """Returns the journal entry as a creation form.

        :param csrf_token: The CSRF token.
        :param next_uri: The next URI.
        :return: The journal entry as a creation form.
        """
        return self.__form(csrf_token, next_uri, is_update=False)

    def update_form(self, csrf_token: str, next_uri: str) -> dict[str, str]:
        """Returns the journal entry as an update form.

        :param csrf_token: The CSRF token.
        :param next_uri: The next URI.
        :return: The journal entry as an update form.
        """
        return self.__form(csrf_token, next_uri, is_update=True)

    def __form(self, csrf_token: str, next_uri: str, is_update: bool = False) \
            -> dict[str, str]:
        """Returns the journal entry as a form.

        :param csrf_token: The CSRF token.
        :param next_uri: The next URI.
        :param is_update: True for an update operation, or False otherwise
        :return: The journal entry as a form.
        """
        journal_entry_date: date = date.today() - timedelta(days=self.days)
        form: dict[str, str] = {"csrf_token": csrf_token,
                                "next": next_uri,
                                "date": journal_entry_date.isoformat()}
        for i in range(len(self.currencies)):
            form.update(self.currencies[i].form(i + 1, is_update))
        if self.note is not None:
            form["note"] = self.note
        return form


class BaseTestData(ABC):
    """The base test data."""

    def __init__(self, app: Flask, username: str):
        """Constructs the test data.

        :param app: The Flask application.
        :param username: The username.
        """
        self._app: Flask = app
        with self._app.app_context():
            current_user: User | None = User.query\
                .filter(User.username == username).first()
            assert current_user is not None
            self.__current_user_id: int = current_user.id
            self.__journal_entries: list[dict[str, t.Any]] = []
            self.__line_items: list[dict[str, t.Any]] = []
            self._init_data()

    @abstractmethod
    def _init_data(self) -> None:
        """Initializes the test data.

        :return: None
        """

    def populate(self) -> None:
        """Populates the data into the database.

        :return: None
        """
        from accounting.models import JournalEntry, JournalEntryLineItem
        with self._app.app_context():
            db.session.execute(sa.insert(JournalEntry), self.__journal_entries)
            db.session.execute(sa.insert(JournalEntryLineItem),
                               self.__line_items)
            db.session.commit()

    @staticmethod
    def _couple(description: str, amount: str, debit: str, credit: str) \
            -> tuple[JournalEntryLineItemData, JournalEntryLineItemData]:
        """Returns a couple of debit-credit line items.

        :param description: The description.
        :param amount: The amount.
        :param debit: The debit account code.
        :param credit: The credit account code.
        :return: The debit line item and credit line item.
        """
        return JournalEntryLineItemData(debit, description, amount),\
            JournalEntryLineItemData(credit, description, amount)

    def _add_journal_entry(self, journal_entry_data: JournalEntryData) -> None:
        """Adds a journal entry.

        :param journal_entry_data: The journal entry data.
        :return: None.
        """
        from accounting.models import Account
        existing_j_id: set[int] = {x["id"] for x in self.__journal_entries}
        existing_l_id: set[int] = {x["id"] for x in self.__line_items}
        journal_entry_data.id = self.__new_id(existing_j_id)
        j_date: date = date.today() - timedelta(days=journal_entry_data.days)
        self.__journal_entries.append(
            {"id": journal_entry_data.id,
             "date": j_date,
             "no": self.__next_j_no(j_date),
             "note": journal_entry_data.note,
             "created_by_id": self.__current_user_id,
             "updated_by_id": self.__current_user_id})
        debit_no: int = 0
        credit_no: int = 0
        for currency in journal_entry_data.currencies:
            for line_item in currency.debit:
                account: Account | None \
                    = Account.find_by_code(line_item.account)
                assert account is not None
                debit_no = debit_no + 1
                line_item.id = self.__new_id(existing_l_id)
                data: dict[str, t.Any] \
                    = {"id": line_item.id,
                       "journal_entry_id": journal_entry_data.id,
                       "is_debit": True,
                       "no": debit_no,
                       "account_id": account.id,
                       "currency_code": currency.code,
                       "description": line_item.description,
                       "amount": line_item.amount}
                if line_item.original_line_item is not None:
                    data["original_line_item_id"] \
                        = line_item.original_line_item.id
                self.__line_items.append(data)
            for line_item in currency.credit:
                account: Account | None \
                    = Account.find_by_code(line_item.account)
                assert account is not None
                credit_no = credit_no + 1
                line_item.id = self.__new_id(existing_l_id)
                data: dict[str, t.Any] \
                    = {"id": line_item.id,
                       "journal_entry_id": journal_entry_data.id,
                       "is_debit": False,
                       "no": credit_no,
                       "account_id": account.id,
                       "currency_code": currency.code,
                       "description": line_item.description,
                       "amount": line_item.amount}
                if line_item.original_line_item is not None:
                    data["original_line_item_id"] \
                        = line_item.original_line_item.id
                self.__line_items.append(data)

    @staticmethod
    def __new_id(existing_id: set[int]) -> int:
        """Generates and returns a new random unique ID.

        :param existing_id: The existing ID.
        :return: The newly-generated random unique ID.
        """
        while True:
            obj_id: int = 100000000 + randbelow(900000000)
            if obj_id not in existing_id:
                existing_id.add(obj_id)
                return obj_id

    def __next_j_no(self, j_date: date) -> int:
        """Returns the next journal entry number in a day.

        :param j_date: The journal entry date.
        :return: The next journal entry number.
        """
        existing: set[int] = {x["no"] for x in self.__journal_entries
                              if x["date"] == j_date}
        return 1 if len(existing) == 0 else max(existing) + 1

    def _add_simple_journal_entry(
            self, days: int, currency: str, description: str, amount: str,
            debit: str, credit: str) \
            -> tuple[JournalEntryLineItemData, JournalEntryLineItemData]:
        """Adds a simple journal entry.

        :param days: The number of days before today.
        :param currency: The currency code.
        :param description: The description.
        :param amount: The amount.
        :param debit: The debit account code.
        :param credit: The credit account code.
        :return: The debit line item and credit line item.
        """
        debit_item, credit_item = self._couple(
            description, amount, debit, credit)
        self._add_journal_entry(JournalEntryData(
            days, [JournalEntryCurrencyData(
                currency, [debit_item], [credit_item])]))
        return debit_item, credit_item



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/25

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The pagination utilities.

This module should not import any other module from the application.

"""
import typing as t
from urllib.parse import urlparse, parse_qsl, urlencode, urlunparse, \
    ParseResult

from flask import request
from werkzeug.routing import RequestRedirect

from accounting.locale import pgettext


class Link:
    """A link."""

    def __init__(self, text: str, uri: str | None = None,
                 is_current: bool = False, is_for_mobile: bool = False):
        """Constructs the link.

        :param text: The link text.
        :param uri: The link URI, or None if there is no link.
        :param is_current: True if the page is the current page, or False
            otherwise.
        :param is_for_mobile: True if the page should be shown on small
            screens, or False otherwise.
        """
        self.text: str = text
        """The link text"""
        self.uri: str | None = uri
        """The link URI, or None if there is no link."""
        self.is_current: bool = is_current
        """Whether the link is the current page."""
        self.is_for_mobile: bool = is_for_mobile
        """Whether the link should be shown on mobile screens."""


class Redirection(RequestRedirect):
    """The redirection."""
    code = 302
    """The HTTP code."""


DEFAULT_PAGE_SIZE: int = 10
"""The default page size."""

T = t.TypeVar("T")


class Pagination(t.Generic[T]):
    """The pagination utility."""

    def __init__(self, items: list[T], is_reversed: bool = False):
        """Constructs the pagination.

        :param items: The items.
        :param is_reversed: True if the default page is the last page, or False
            otherwise.
        :raise Redirection: When the pagination parameters are malformed.
        """
        pagination: AbstractPagination[T] = EmptyPagination[T]() \
            if len(items) == 0 \
            else NonEmptyPagination[T](items, is_reversed)
        self.is_paged: bool = pagination.is_paged
        """Whether there should be pagination."""
        self.list: list[T] = pagination.list
        """The items shown in the list"""
        self.pages: list[Link] = pagination.pages
        """The pages."""
        self.page_size: int = pagination.page_size
        """The number of items in a page."""
        self.page_size_options: list[Link] = pagination.page_size_options
        """The options to the number of items in a page."""


class AbstractPagination(t.Generic[T]):
    """An abstract pagination."""

    def __init__(self):
        """Constructs an empty pagination."""
        self.page_size: int = DEFAULT_PAGE_SIZE
        """The number of items in a page."""
        self.is_paged: bool = False
        """Whether there should be pagination."""
        self.list: list[T] = []
        """The items shown in the list"""
        self.pages: list[Link] = []
        """The pages."""
        self.page_size_options: list[Link] = []
        """The options to the number of items in a page."""


class EmptyPagination(AbstractPagination[T]):
    """The pagination from empty data."""
    pass


class NonEmptyPagination(AbstractPagination[T]):
    """The pagination with real data."""
    PAGE_SIZE_OPTION_VALUES: list[int] = [10, 100, 200]
    """The page size options."""

    def __init__(self, items: list[T], is_reversed: bool = False):
        """Constructs the pagination.

        :param items: The items.
        :param is_reversed: True if the default page is the last page, or False
            otherwise.
        :raise Redirection: When the pagination parameters are malformed.
        """
        super().__init__()
        self.__current_uri: str = request.full_path if request.query_string \
            else request.path
        """The current URI."""
        self.__is_reversed: bool = is_reversed
        """Whether the default page is the last page."""
        self.page_size = self.__get_page_size()
        self.__total_pages: int = int((len(items) - 1) / self.page_size) + 1
        """The total number of pages."""
        self.is_paged = self.__total_pages > 1
        self.__default_page_no: int = self.__total_pages \
            if self.__is_reversed else 1
        """The default page number."""
        self.__page_no: int = self.__get_page_no()
        """The current page number."""
        lower_bound: int = (self.__page_no - 1) * self.page_size
        upper_bound: int = lower_bound + self.page_size
        if upper_bound > len(items):
            upper_bound = len(items)
        self.list = items[lower_bound:upper_bound]
        self.pages = self.__get_pages()
        self.page_size_options = self.__get_page_size_options()

    def __get_page_size(self) -> int:
        """Returns the page size.

        :return: The page size.
        :raise Redirection: When the page size is malformed.
        """
        if "page-size" not in request.args:
            return DEFAULT_PAGE_SIZE
        try:
            page_size: int = int(request.args["page-size"])
        except ValueError:
            raise Redirection(self.__uri_set("page-size", None))
        if page_size == DEFAULT_PAGE_SIZE or page_size < 1:
            raise Redirection(self.__uri_set("page-size", None))
        return page_size

    def __get_page_no(self) -> int:
        """Returns the page number.

        :return: The page number.
        :raise Redirection: When the page number is malformed.
        """
        if "page-no" not in request.args:
            return self.__default_page_no
        try:
            page_no: int = int(request.args["page-no"])
        except ValueError:
            raise Redirection(self.__uri_set("page-no", None))
        if page_no == self.__default_page_no:
            raise Redirection(self.__uri_set("page-no", None))
        if page_no < 1:
            if not self.__is_reversed:
                raise Redirection(self.__uri_set("page-no", None))
            raise Redirection(self.__uri_set("page-no", "1"))
        if page_no > self.__total_pages:
            if self.__is_reversed:
                raise Redirection(self.__uri_set("page-no", None))
            raise Redirection(self.__uri_set("page-no",
                                             str(self.__total_pages)))
        return page_no

    def __get_pages(self) -> list[Link]:
        """Returns the page links in the pagination navigation.

        :return: The page links in the pagination navigation.
        """
        if not self.is_paged:
            return []
        uri: str | None
        links: list[Link] = []

        # The previous page.
        uri = None if self.__page_no == 1 \
            else self.__uri_page(self.__page_no - 1)
        links.append(Link(pgettext("Pagination|", "Previous"), uri,
                          is_for_mobile=True))

        # The first page.
        if self.__page_no > 1:
            links.append(Link("1", self.__uri_page(1)))

        # The eclipse of the previous pages.
        if self.__page_no - 3 == 2:
            links.append(Link(str(self.__page_no - 3),
                              self.__uri_page(self.__page_no - 3)))
        elif self.__page_no - 3 > 2:
            links.append(Link("…"))

        # The previous two pages.
        if self.__page_no - 2 > 1:
            links.append(Link(str(self.__page_no - 2),
                              self.__uri_page(self.__page_no - 2)))
        if self.__page_no - 1 > 1:
            links.append(Link(str(self.__page_no - 1),
                              self.__uri_page(self.__page_no - 1)))

        # The current page.
        links.append(Link(str(self.__page_no), self.__uri_page(self.__page_no),
                          is_current=True))

        # The next two pages.
        if self.__page_no + 1 < self.__total_pages:
            links.append(Link(str(self.__page_no + 1),
                              self.__uri_page(self.__page_no + 1)))
        if self.__page_no + 2 < self.__total_pages:
            links.append(Link(str(self.__page_no + 2),
                              self.__uri_page(self.__page_no + 2)))

        # The eclipse of the next pages.
        if self.__page_no + 3 == self.__total_pages - 1:
            links.append(Link(str(self.__page_no + 3),
                              self.__uri_page(self.__page_no + 3)))
        elif self.__page_no + 3 < self.__total_pages - 1:
            links.append(Link("…"))

        # The last page.
        if self.__page_no < self.__total_pages:
            links.append(Link(str(self.__total_pages),
                              self.__uri_page(self.__total_pages)))

        # The next page.
        uri = None if self.__page_no == self.__total_pages \
            else self.__uri_page(self.__page_no + 1)
        links.append(Link(pgettext("Pagination|", "Next"), uri,
                          is_for_mobile=True))

        return links

    def __uri_page(self, page_no: int) -> str:
        """Returns the URI of a page.

        :param page_no: The page number.
        :return: The URI of the page.
        """
        if page_no == self.__page_no:
            return self.__current_uri
        if page_no == self.__default_page_no:
            return self.__uri_set("page-no", None)
        return self.__uri_set("page-no", str(page_no))

    def __get_page_size_options(self) -> list[Link]:
        """Returns the page size options.

        :return: The page size options.
        """
        if not self.is_paged:
            return []
        return [Link(str(x), self.__uri_size(x),
                     is_current=x == self.page_size)
                for x in self.PAGE_SIZE_OPTION_VALUES]

    def __uri_size(self, page_size: int) -> str:
        """Returns the URI of a page size.

        :param page_size: The page size.
        :return: The URI of the page size.
        """
        if page_size == self.page_size:
            return self.__current_uri
        if page_size == DEFAULT_PAGE_SIZE:
            return self.__uri_set("page-size", None)
        return self.__uri_set("page-size", str(page_size))

    def __uri_set(self, name: str, value: str | None) -> str:
        """Raises current URI with a parameter set.

        :param name: The name of the parameter.
        :param value: The value, or None to remove the parameter.
        :return: The URI with the parameter set.
        """
        uri_p: ParseResult = urlparse(self.__current_uri)
        params: list[tuple[str, str]] = parse_qsl(uri_p.query)

        # Try to keep the position of the parameter.
        i: int = 0
        is_found: bool = False
        while i < len(params):
            if params[i][0] == name:
                if is_found or value is None:
                    params = params[:i] + params[i + 1:]
                    continue
                params[i] = (name, value)
                is_found = True
            i = i + 1
        if not is_found and value is not None:
            params.append((name, value))

        parts: list[str] = list(uri_p)
        parts[4] = urlencode(params)
        return urlunparse(parts)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/4

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The journal entry types.

"""
from enum import Enum


class JournalEntryType(Enum):
    """The journal entry types."""
    CASH_RECEIPT: str = "receipt"
    """The cash receipt journal entry."""
    CASH_DISBURSEMENT: str = "disbursement"
    """The cash disbursement journal entry."""
    TRANSFER: str = "transfer"
    """The transfer journal entry."""



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/22

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The forms.

"""
import re

from flask_babel import LazyString
from flask_wtf import FlaskForm
from wtforms import StringField, ValidationError
from wtforms.validators import DataRequired

from accounting import db
from accounting.locale import lazy_gettext
from accounting.models import Currency, Account


ACCOUNT_REQUIRED: DataRequired = DataRequired(
    lazy_gettext("Please select the account."))
"""The validator to check if the account code is empty."""


class CurrencyExists:
    """The validator to check if the account exists."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        if field.data is None:
            return
        if db.session.get(Currency, field.data) is None:
            raise ValidationError(lazy_gettext(
                "The currency does not exist."))


class AccountExists:
    """The validator to check if the account exists."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        if field.data is None:
            return
        if Account.find_by_code(field.data) is None:
            raise ValidationError(lazy_gettext(
                "The account does not exist."))


class IsDebitAccount:
    """The validator to check if the account is for debit line items."""

    def __init__(self, message: str | LazyString):
        """Constructs the validator.

        :param message: The error message.
        """
        self.__message: str | LazyString = message

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        if field.data is None:
            return
        if re.match(r"^(?:[1235689]|7[5678])", field.data) \
                and not field.data.startswith("3351-") \
                and not field.data.startswith("3353-"):
            return
        raise ValidationError(self.__message)


class IsCreditAccount:
    """The validator to check if the account is for credit line items."""

    def __init__(self, message: str | LazyString):
        """Constructs the validator.

        :param message: The error message.
        """
        self.__message: str | LazyString = message

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        if field.data is None:
            return
        if re.match(r"^(?:[123489]|7[1234])", field.data) \
                and not field.data.startswith("3351-") \
                and not field.data.startswith("3353-"):
            return
        raise ValidationError(self.__message)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/8

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The search.

"""
from datetime import datetime
from decimal import Decimal

import sqlalchemy as sa
from flask import Response, render_template, request
from sqlalchemy.orm import selectinload

from accounting.locale import gettext
from accounting.models import Currency, CurrencyL10n, Account, AccountL10n, \
    JournalEntry, JournalEntryLineItem
from accounting.report.utils.base_page_params import BasePageParams
from accounting.report.utils.base_report import BaseReport
from accounting.report.utils.csv_export import csv_download
from accounting.report.utils.report_chooser import ReportChooser
from accounting.report.utils.report_type import ReportType
from accounting.utils.cast import be
from accounting.utils.pagination import Pagination
from accounting.utils.query import parse_query_keywords
from .journal import get_csv_rows


class LineItemCollector:
    """The line item collector."""

    def __init__(self):
        """Constructs the line item collector."""
        self.line_items: list[JournalEntryLineItem] = self.__query_line_items()
        """The line items."""

    def __query_line_items(self) -> list[JournalEntryLineItem]:
        """Queries and returns the line items.

        :return: The line items.
        """
        keywords: list[str] = parse_query_keywords(request.args.get("q"))
        if len(keywords) == 0:
            return []
        conditions: list[sa.BinaryExpression] = []
        for k in keywords:
            sub_conditions: list[sa.BinaryExpression] \
                = [JournalEntryLineItem.description.icontains(k),
                   JournalEntryLineItem.account_id.in_(
                       self.__get_account_condition(k)),
                   JournalEntryLineItem.currency_code.in_(
                       self.__get_currency_condition(k)),
                   JournalEntryLineItem.journal_entry_id.in_(
                       self.__get_journal_entry_condition(k))]
            try:
                sub_conditions.append(
                    JournalEntryLineItem.amount == Decimal(k))
            except ArithmeticError:
                pass
            conditions.append(sa.or_(*sub_conditions))
        return JournalEntryLineItem.query.join(JournalEntry)\
            .filter(*conditions)\
            .order_by(JournalEntry.date,
                      JournalEntry.no,
                      JournalEntryLineItem.is_debit,
                      JournalEntryLineItem.no)\
            .options(selectinload(JournalEntryLineItem.account),
                     selectinload(JournalEntryLineItem.currency),
                     selectinload(JournalEntryLineItem.journal_entry)).all()

    @staticmethod
    def __get_account_condition(k: str) -> sa.Select:
        """Composes and returns the condition to filter the account.

        :param k: The keyword.
        :return: The condition to filter the account.
        """
        code: sa.BinaryExpression = Account.base_code + "-" \
            + sa.func.substr("000" + sa.cast(Account.no, sa.String),
                             sa.func.char_length(sa.cast(Account.no,
                                                         sa.String)) + 1)
        select_l10n: sa.Select = sa.select(AccountL10n.account_id)\
            .filter(AccountL10n.title.icontains(k))
        conditions: list[sa.BinaryExpression] \
            = [Account.base_code.contains(k),
               Account.title_l10n.icontains(k),
               code.contains(k),
               Account.id.in_(select_l10n)]
        if k in gettext("Needs Offset"):
            conditions.append(Account.is_need_offset)
        return sa.select(Account.id).filter(sa.or_(*conditions))

    @staticmethod
    def __get_currency_condition(k: str) -> sa.Select:
        """Composes and returns the condition to filter the currency.

        :param k: The keyword.
        :return: The condition to filter the currency.
        """
        select_l10n: sa.Select = sa.select(CurrencyL10n.currency_code)\
            .filter(CurrencyL10n.name.icontains(k))
        return sa.select(Currency.code).filter(
            sa.or_(Currency.code.icontains(k),
                   Currency.name_l10n.icontains(k),
                   Currency.code.in_(select_l10n)))

    @staticmethod
    def __get_journal_entry_condition(k: str) -> sa.Select:
        """Composes and returns the condition to filter the journal entry.

        :param k: The keyword.
        :return: The condition to filter the journal entry.
        """
        conditions: list[sa.BinaryExpression] \
            = [JournalEntry.note.icontains(k)]
        journal_entry_date: datetime
        try:
            journal_entry_date = datetime.strptime(k, "%Y")
            conditions.append(
                be(sa.extract("year", JournalEntry.date)
                   == journal_entry_date.year))
        except ValueError:
            pass
        try:
            journal_entry_date = datetime.strptime(k, "%Y/%m")
            conditions.append(sa.and_(
                sa.extract("year", JournalEntry.date)
                == journal_entry_date.year,
                sa.extract("month", JournalEntry.date)
                == journal_entry_date.month))
        except ValueError:
            pass
        try:
            journal_entry_date = datetime.strptime(f"2000/{k}", "%Y/%m/%d")
            conditions.append(sa.and_(
                sa.extract("month", JournalEntry.date)
                == journal_entry_date.month,
                sa.extract("day", JournalEntry.date)
                == journal_entry_date.day))
        except ValueError:
            pass
        try:
            journal_entry_date = datetime.strptime(k, "%Y/%m/%d")
            conditions.append(sa.and_(
                sa.extract("year", JournalEntry.date)
                == journal_entry_date.year,
                sa.extract("month", JournalEntry.date)
                == journal_entry_date.month,
                sa.extract("day", JournalEntry.date)
                == journal_entry_date.day))
        except ValueError:
            pass
        return sa.select(JournalEntry.id).filter(sa.or_(*conditions))


class PageParams(BasePageParams):
    """The HTML page parameters."""

    def __init__(self, pagination: Pagination[JournalEntryLineItem],
                 line_items: list[JournalEntryLineItem]):
        """Constructs the HTML page parameters.

        :param line_items: The search result line items.
        """
        self.pagination: Pagination[JournalEntryLineItem] = pagination
        """The pagination."""
        self.line_items: list[JournalEntryLineItem] = line_items
        """The line items."""

    @property
    def has_data(self) -> bool:
        """Returns whether there is any data on the page.

        :return: True if there is any data, or False otherwise.
        """
        return len(self.line_items) > 0

    @property
    def report_chooser(self) -> ReportChooser:
        """Returns the report chooser.

        :return: The report chooser.
        """
        return ReportChooser(ReportType.SEARCH)


class Search(BaseReport):
    """The search."""

    def __init__(self):
        """Constructs a search."""
        self.__line_items: list[JournalEntryLineItem] \
            = LineItemCollector().line_items
        """The line items."""

    def csv(self) -> Response:
        """Returns the report as CSV for download.

        :return: The response of the report for download.
        """
        filename: str = "search-{q}.csv".format(q=request.args["q"])
        return csv_download(filename, get_csv_rows(self.__line_items))

    def html(self) -> str:
        """Composes and returns the report as HTML.

        :return: The report as HTML.
        """
        pagination: Pagination[JournalEntryLineItem] \
            = Pagination[JournalEntryLineItem](self.__line_items,
                                               is_reversed=True)
        params: PageParams = PageParams(pagination=pagination,
                                        line_items=pagination.list)
        return render_template("accounting/report/search.html",
                               report=params)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/7

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The unapplied original line item utilities.

"""
import sqlalchemy as sa

from accounting import db
from accounting.models import Account, JournalEntryLineItem
from accounting.utils.cast import be
from accounting.utils.offset_alias import offset_alias


def get_accounts_with_unapplied() -> list[Account]:
    """Returns the accounts with unapplied original line items.

    :return: The accounts with unapplied original line items.
    """
    offset: sa.Alias = offset_alias()
    net_balance: sa.Label \
        = (JournalEntryLineItem.amount
           + sa.func.sum(sa.case(
                (be(offset.c.is_debit == JournalEntryLineItem.is_debit),
                 offset.c.amount),
                else_=-offset.c.amount))).label("net_balance")
    select_unapplied: sa.Select \
        = sa.select(JournalEntryLineItem.id)\
        .join(Account)\
        .join(offset, be(JournalEntryLineItem.id
                         == offset.c.original_line_item_id),
              isouter=True)\
        .filter(Account.is_need_offset,
                sa.or_(sa.and_(Account.base_code.startswith("2"),
                               sa.not_(JournalEntryLineItem.is_debit)),
                       sa.and_(Account.base_code.startswith("1"),
                               JournalEntryLineItem.is_debit)))\
        .group_by(JournalEntryLineItem.id)\
        .having(sa.or_(sa.func.count(offset.c.id) == 0, net_balance != 0))

    count_func: sa.Label \
        = sa.func.count(JournalEntryLineItem.id).label("count")
    select: sa.Select = sa.select(Account.id, count_func)\
        .join(JournalEntryLineItem, isouter=True)\
        .filter(JournalEntryLineItem.id.in_(select_unapplied))\
        .group_by(Account.id)\
        .having(count_func > 0)
    counts: dict[int, int] \
        = {x.id: x.count for x in db.session.execute(select)}
    accounts: list[Account] = Account.query.filter(Account.id.in_(counts))\
        .order_by(Account.base_code, Account.no).all()
    for account in accounts:
        account.count = counts[account.id]
    return accounts



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/27

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The description editor.

"""
import re
import typing as t

import sqlalchemy as sa

from accounting import db
from accounting.models import Account, JournalEntryLineItem
from accounting.utils.options import options, Recurring


class DescriptionAccount:
    """An account for a description tag."""

    def __init__(self, account: Account, freq: int):
        """Constructs an account for a description tag.

        :param account: The account.
        :param freq: The frequency of the tag with the account.
        """
        self.__account: Account = account
        """The account."""
        self.id: int = account.id
        """The account ID."""
        self.code: str = account.code
        """The account code."""
        self.is_need_offset: bool = account.is_need_offset
        """Whether the journal entry line items of this account need offset."""
        self.freq: int = freq
        """The frequency of the tag with the account."""

    def __str__(self) -> str:
        """Returns the string representation of the account.

        :return: The string representation of the account.
        """
        return str(self.__account)

    def add_freq(self, freq: int) -> None:
        """Adds the frequency of an account.

        :param freq: The frequency of the tag name with the account.
        :return: None.
        """
        self.freq = self.freq + freq


class DescriptionTag:
    """A description tag."""

    def __init__(self, name: str):
        """Constructs a description tag.

        :param name: The tag name.
        """
        self.name: str = name
        """The tag name."""
        self.__account_dict: dict[int, DescriptionAccount] = {}
        """The accounts that come with the tag, in the order of their
        frequency."""
        self.freq: int = 0
        """The frequency of the tag."""

    def __str__(self) -> str:
        """Returns the string representation of the tag.

        :return: The string representation of the tag.
        """
        return self.name

    def add_account(self, account: Account, freq: int):
        """Adds an account.

        :param account: The associated account.
        :param freq: The frequency of the tag name with the account.
        :return: None.
        """
        self.__account_dict[account.id] = DescriptionAccount(account, freq)
        self.freq = self.freq + freq

    @property
    def accounts(self) -> list[DescriptionAccount]:
        """Returns the accounts by the order of their frequencies.

        :return: The accounts by the order of their frequencies.
        """
        return sorted(self.__account_dict.values(), key=lambda x: -x.freq)

    @property
    def account_codes(self) -> list[str]:
        """Returns the account codes by the order of their frequencies.

        :return: The account codes by the order of their frequencies.
        """
        return [x.code for x in self.accounts]


class DescriptionType:
    """A description type"""

    def __init__(self, type_id: t.Literal["general", "travel", "bus"]):
        """Constructs a description type.

        :param type_id: The type ID, either "general", "travel", or "bus".
        """
        self.id: t.Literal["general", "travel", "bus"] = type_id
        """The type ID."""
        self.__tag_dict: dict[str, DescriptionTag] = {}
        """A dictionary from the tag name to their corresponding tag."""

    def add_tag(self, name: str, account: Account, freq: int) -> None:
        """Adds a tag.

        :param name: The tag name.
        :param account: The associated account.
        :param freq: The frequency of the tag name with the account.
        :return: None.
        """
        if name not in self.__tag_dict:
            self.__tag_dict[name] = DescriptionTag(name)
        self.__tag_dict[name].add_account(account, freq)

    @property
    def tags(self) -> list[DescriptionTag]:
        """Returns the tags by the order of their frequencies.

        :return: The tags by the order of their frequencies.
        """
        return sorted(self.__tag_dict.values(), key=lambda x: -x.freq)


class DescriptionRecurring:
    """A recurring transaction."""

    def __init__(self, name: str, account: Account, description_template: str):
        """Constructs a recurring transaction.

        :param name: The name.
        :param description_template: The description template.
        :param account: The account.
        """
        self.name: str = name
        self.account: DescriptionAccount = DescriptionAccount(account, 0)
        self.description_template: str = description_template

    @property
    def account_codes(self) -> list[str]:
        """Returns the account codes by the order of their frequencies.

        :return: The account codes by the order of their frequencies.
        """
        return [self.account.code]


class DescriptionDebitCredit:
    """The description on debit or credit."""

    def __init__(self, debit_credit: t.Literal["debit", "credit"]):
        """Constructs the description on debit or credit.

        :param debit_credit: Either "debit" or "credit".
        """
        self.debit_credit: t.Literal["debit", "credit"] = debit_credit
        """Either debit or credit."""
        self.general: DescriptionType = DescriptionType("general")
        """The general tags."""
        self.travel: DescriptionType = DescriptionType("travel")
        """The travel tags."""
        self.bus: DescriptionType = DescriptionType("bus")
        """The bus tags."""
        self.__type_dict: dict[t.Literal["general", "travel", "bus"],
                               DescriptionType] \
            = {x.id: x for x in {self.general, self.travel, self.bus}}
        """A dictionary from the type ID to the corresponding tags."""
        self.recurring: list[DescriptionRecurring] = []
        """The recurring transactions."""

    def add_tag(self, tag_type: t.Literal["general", "travel", "bus"],
                name: str, account: Account, freq: int) -> None:
        """Adds a tag.

        :param tag_type: The tag type, either "general", "travel", or "bus".
        :param name: The name.
        :param account: The associated account.
        :param freq: The frequency of the tag name with the account.
        :return: None.
        """
        self.__type_dict[tag_type].add_tag(name, account, freq)

    @property
    def accounts(self) -> list[DescriptionAccount]:
        """Returns the suggested accounts of all tags in the description editor
        in debit or credit, in their frequency order.

        :return: The suggested accounts of all tags, in their frequency order.
        """
        accounts: dict[int, DescriptionAccount] = {}
        freq: dict[int, int] = {}
        for tag_type in self.__type_dict.values():
            for tag in tag_type.tags:
                for account in tag.accounts:
                    accounts[account.id] = account
                    if account.id not in freq:
                        freq[account.id] = 0
                    freq[account.id] \
                        = freq[account.id] + account.freq
        for recurring in self.recurring:
            accounts[recurring.account.id] = recurring.account
            if recurring.account.id not in freq:
                freq[recurring.account.id] = 0
        return [accounts[y] for y in sorted(freq.keys(),
                                            key=lambda x: -freq[x])]


class DescriptionEditor:
    """The description editor."""

    def __init__(self):
        """Constructs the description editor."""
        self.debit: DescriptionDebitCredit = DescriptionDebitCredit("debit")
        """The debit tags."""
        self.credit: DescriptionDebitCredit = DescriptionDebitCredit("credit")
        """The credit tags."""
        self.__init_tags()
        self.__init_recurring()

    def __init_tags(self):
        """Initializes the tags.

        :return: None.
        """
        debit_credit: sa.Label = sa.case(
            (JournalEntryLineItem.is_debit, "debit"),
            else_="credit").label("debit_credit")
        tag_type: sa.Label = sa.case(
            (JournalEntryLineItem.description.like("_%—_%—_%→_%"), "bus"),
            (sa.or_(JournalEntryLineItem.description.like("_%—_%→_%"),
                    JournalEntryLineItem.description.like("_%—_%↔_%")),
             "travel"),
            else_="general").label("tag_type")
        tag: sa.Label = get_prefix(JournalEntryLineItem.description, "—")\
            .label("tag")
        select: sa.Select = sa.Select(debit_credit, tag_type, tag,
                                      JournalEntryLineItem.account_id,
                                      sa.func.count().label("freq"))\
            .filter(JournalEntryLineItem.description.is_not(None),
                    JournalEntryLineItem.description.like("_%—_%"),
                    JournalEntryLineItem.original_line_item_id.is_(None))\
            .group_by(debit_credit, tag_type, tag,
                      JournalEntryLineItem.account_id)
        result: list[sa.Row] = db.session.execute(select).all()
        accounts: dict[int, Account] \
            = {x.id: x for x in Account.query
               .filter(Account.id.in_({x.account_id for x in result})).all()}
        debit_credit_dict: dict[t.Literal["debit", "credit"],
                                DescriptionDebitCredit] \
            = {x.debit_credit: x for x in {self.debit, self.credit}}
        for row in result:
            debit_credit_dict[row.debit_credit].add_tag(
                row.tag_type, row.tag, accounts[row.account_id], row.freq)

    def __init_recurring(self) -> None:
        """Initializes the recurring transactions.

        :return: None.
        """
        recurring: Recurring = options.recurring
        accounts: dict[str, Account] \
            = self.__get_accounts(recurring.codes)
        self.debit.recurring \
            = [DescriptionRecurring(x.name, accounts[x.account_code],
                                    x.description_template)
               for x in recurring.expenses]
        self.credit.recurring \
            = [DescriptionRecurring(x.name, accounts[x.account_code],
                                    x.description_template)
               for x in recurring.incomes]

    @staticmethod
    def __get_accounts(codes: set[str]) -> dict[str, Account]:
        """Finds and returns the accounts by codes.

        :param codes: The account codes.
        :return: The account.
        """
        if len(codes) == 0:
            return {}

        def get_condition(code0: str) -> sa.BinaryExpression:
            m: re.Match = re.match(r"^(\d{4})-(\d{3})$", code0)
            assert m is not None,\
                f"Malformed account code \"{code0}\" for regular transactions."
            return sa.and_(Account.base_code == m.group(1),
                           Account.no == int(m.group(2)))

        conditions: list[sa.BinaryExpression] \
            = [get_condition(x) for x in codes]
        accounts: dict[str, Account] \
            = {x.code: x for x in
               Account.query.filter(sa.or_(*conditions)).all()}
        for code in codes:
            assert code in accounts,\
                f"Unknown account \"{code}\" for regular transactions."
        return accounts


def get_prefix(string: str | sa.Column, separator: str | sa.Column) \
        -> sa.Function:
    """Returns the SQL function to find the prefix of a string.

    :param string: The string.
    :param separator: The separator.
    :return: The position of the substring, starting from 1.
    """
    return sa.func.substr(string, 0, get_position(string, separator))


def get_position(string: str | sa.Column, substring: str | sa.Column) \
        -> sa.Function:
    """Returns the SQL function to find the position of a substring.

    :param string: The string.
    :param substring: The substring.
    :return: The position of the substring, starting from 1.
    """
    if db.engine.name == "postgresql":
        return sa.func.strpos(string, substring)
    return sa.func.instr(string, substring)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/9

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The utilities to get the ledger URL.

"""
from flask import url_for

from accounting.models import Currency, Account
from accounting.report.period import Period
from accounting.template_globals import default_currency_code
from accounting.utils.current_account import CurrentAccount
from accounting.utils.options import options


def journal_url(period: Period) \
        -> str:
    """Returns the URL of a journal.

    :param period: The period.
    :return: The URL of the journal.
    """
    if period.is_default:
        return url_for("accounting-report.journal-default")
    return url_for("accounting-report.journal", period=period)


def ledger_url(currency: Currency, account: Account, period: Period) \
        -> str:
    """Returns the URL of a ledger.

    :param currency: The currency.
    :param account: The account.
    :param period: The period.
    :return: The URL of the ledger.
    """
    if currency.code == default_currency_code() \
            and account.code == Account.CASH_CODE \
            and period.is_default:
        return url_for("accounting-report.ledger-default")
    return url_for("accounting-report.ledger",
                   currency=currency, account=account,
                   period=period)


def income_expenses_url(currency: Currency, account: CurrentAccount,
                        period: Period) -> str:
    """Returns the URL of an income and expenses log.

    :param currency: The currency.
    :param account: The account.
    :param period: The period.
    :return: The URL of the income and expenses log.
    """
    if currency.code == default_currency_code() \
            and account.code == options.default_ie_account_code \
            and period.is_default:
        return url_for("accounting-report.default")
    return url_for("accounting-report.income-expenses",
                   currency=currency, account=account,
                   period=period)


def trial_balance_url(currency: Currency, period: Period) -> str:
    """Returns the URL of a trial balance.

    :param currency: The currency.
    :param period: The period.
    :return: The URL of the trial balance.
    """
    if currency.code == default_currency_code() and period.is_default:
        return url_for("accounting-report.trial-balance-default")
    return url_for("accounting-report.trial-balance",
                   currency=currency, period=period)


def income_statement_url(currency: Currency, period: Period) -> str:
    """Returns the URL of an income statement.

    :param currency: The currency.
    :param period: The period.
    :return: The URL of the income statement.
    """
    if currency.code == default_currency_code() and period.is_default:
        return url_for("accounting-report.income-statement-default")
    return url_for("accounting-report.income-statement",
                   currency=currency, period=period)


def balance_sheet_url(currency: Currency, period: Period) -> str:
    """Returns the URL of a balance sheet.

    :param currency: The currency.
    :param period: The period.
    :return: The URL of the balance sheet.
    """
    if currency.code == default_currency_code() and period.is_default:
        return url_for("accounting-report.balance-sheet-default")
    return url_for("accounting-report.balance-sheet",
                   currency=currency, period=period)


def unapplied_url(account: Account | None) -> str:
    """Returns the URL of the unapplied original line items.

    :param account: The account, or None to list the accounts with unapplied
        original line items.
    :return: The URL of the unapplied original line items.
    """
    if account is None:
        return url_for("accounting-report.unapplied-default")
    return url_for("accounting-report.unapplied", account=account)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/8

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The forms for the unmatched offset management.

"""
import sqlalchemy as sa
from sqlalchemy.orm import selectinload

from accounting.models import Account, JournalEntry, JournalEntryLineItem
from accounting.utils.unapplied import get_unapplied_original_line_items


class OffsetPair:
    """A pair of an original line item and its offset."""

    def __init__(self, original_line_item: JournalEntryLineItem,
                 offset: JournalEntryLineItem):
        """Constructs a pair of an original line item and its offset.

        :param original_line_item: The original line item.
        :param offset: The offset.
        """
        self.original_line_item: JournalEntryLineItem = original_line_item
        """The original line item."""
        self.offset: JournalEntryLineItem = offset
        """The offset."""


class OffsetMatcher:
    """The offset matcher."""

    def __init__(self, account: Account):
        """Constructs the offset matcher.

        :param account: The account.
        """
        self.account: Account = account
        """The account."""
        self.matched_pairs: list[OffsetPair] = []
        """A list of matched pairs."""
        self.is_having_matches: bool = False
        """Whether there is any matches."""
        self.total: int = 0
        """The total number of unapplied debits or credits."""
        self.unapplied: list[JournalEntryLineItem] = []
        """The unapplied debits or credits."""
        self.unmatched_offsets: list[JournalEntryLineItem] = []
        """The unmatched offsets."""
        self.__find_matches()

    def __find_matches(self) -> None:
        """Finds the matched original line items and their offsets.

        :return: None.
        """
        self.unapplied: list[JournalEntryLineItem] \
            = get_unapplied_original_line_items(self.account)
        self.total = len(self.unapplied)
        if self.total == 0:
            self.is_having_matches = False
            return
        self.unmatched_offsets = self.__get_unmatched_offsets()
        remains: list[JournalEntryLineItem] = self.unmatched_offsets.copy()
        for original_item in self.unapplied:
            offset_candidates: list[JournalEntryLineItem] \
                = [x for x in remains
                   if (x.journal_entry.date > original_item.journal_entry.date
                       or (x.journal_entry.date
                           == original_item.journal_entry.date
                           and x.journal_entry.no
                           > original_item.journal_entry.no))
                   and x.currency_code == original_item.currency_code
                   and x.description == original_item.description
                   and x.amount == original_item.net_balance]
            if len(offset_candidates) == 0:
                continue
            self.matched_pairs.append(
                OffsetPair(original_item, offset_candidates[0]))
            original_item.match = offset_candidates[0]
            offset_candidates[0].match = original_item
            remains.remove(offset_candidates[0])
        self.is_having_matches = len(self.matched_pairs) > 0

    def __get_unmatched_offsets(self) -> list[JournalEntryLineItem]:
        """Returns the unmatched offsets of an account.

        :return: The unmatched offsets of the account.
        """
        return JournalEntryLineItem.query.join(Account).join(JournalEntry)\
            .filter(Account.id == self.account.id,
                    JournalEntryLineItem.original_line_item_id.is_(None),
                    sa.or_(sa.and_(Account.base_code.startswith("2"),
                                   JournalEntryLineItem.is_debit),
                           sa.and_(Account.base_code.startswith("1"),
                                   sa.not_(JournalEntryLineItem.is_debit))))\
            .order_by(JournalEntry.date, JournalEntry.no,
                      JournalEntryLineItem.is_debit, JournalEntryLineItem.no)\
            .options(selectinload(JournalEntryLineItem.currency),
                     selectinload(JournalEntryLineItem.journal_entry)).all()

    @property
    def matches(self) -> int:
        """Returns the number of matches.

        :return: The number of matches.
        """
        return len(self.matched_pairs)

    def match(self) -> None:
        """Matches the original line items with offsets.

        :return: None.
        """
        for pair in self.matched_pairs:
            pair.offset.original_line_item_id = pair.original_line_item.id



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/7

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The unapplied original line items.

"""
from datetime import date
from decimal import Decimal

from flask import render_template, Response

from accounting.locale import gettext
from accounting.models import Account, JournalEntryLineItem
from accounting.report.utils.base_page_params import BasePageParams
from accounting.report.utils.base_report import BaseReport
from accounting.report.utils.csv_export import BaseCSVRow, csv_download
from accounting.report.utils.option_link import OptionLink
from accounting.report.utils.report_chooser import ReportChooser
from accounting.report.utils.report_type import ReportType
from accounting.report.utils.unapplied import get_accounts_with_unapplied
from accounting.report.utils.urls import unapplied_url
from accounting.utils.offset_matcher import OffsetMatcher
from accounting.utils.pagination import Pagination
from accounting.utils.permission import can_edit


class CSVRow(BaseCSVRow):
    """A row in the CSV."""

    def __init__(self, journal_entry_date: str | date, currency: str,
                 description: str | None, amount: str | Decimal,
                 net_balance: str | Decimal):
        """Constructs a row in the CSV.

        :param journal_entry_date: The journal entry date.
        :param currency: The currency.
        :param description: The description.
        :param amount: The amount.
        :param net_balance: The net balance.
        """
        self.date: str | date = journal_entry_date
        """The date."""
        self.currency: str = currency
        """The currency."""
        self.description: str | None = description
        """The description."""
        self.amount: str | Decimal = amount
        """The amount."""
        self.net_balance: str | Decimal = net_balance
        """The net balance."""

    @property
    def values(self) -> list[str | date | Decimal | None]:
        """Returns the values of the row.

        :return: The values of the row.
        """
        return [self.date, self.currency, self.description, self.amount,
                self.net_balance]


class PageParams(BasePageParams):
    """The HTML page parameters."""

    def __init__(self, account: Account,
                 is_mark_matches: bool,
                 pagination: Pagination[JournalEntryLineItem],
                 line_items: list[JournalEntryLineItem]):
        """Constructs the HTML page parameters.

        :param account: The account.
        :param is_mark_matches: Whether to mark the matched offsets.
        :param pagination: The pagination.
        :param line_items: The line items.
        """
        self.account: Account = account
        """The account."""
        self.pagination: Pagination[JournalEntryLineItem] = pagination
        """The pagination."""
        self.line_items: list[JournalEntryLineItem] = line_items
        """The line items."""
        self.is_mark_matches: bool = is_mark_matches
        """Whether to mark the matched offsets."""

    @property
    def has_data(self) -> bool:
        """Returns whether there is any data on the page.

        :return: True if there is any data, or False otherwise.
        """
        return len(self.line_items) > 0

    @property
    def report_chooser(self) -> ReportChooser:
        """Returns the report chooser.

        :return: The report chooser.
        """
        return ReportChooser(ReportType.UNAPPLIED,
                             account=self.account)

    @property
    def account_options(self) -> list[OptionLink]:
        """Returns the account options.

        :return: The account options.
        """
        options: list[OptionLink] = [OptionLink(gettext("Accounts"),
                                                unapplied_url(None),
                                                False)]
        options.extend([OptionLink(str(x),
                                   unapplied_url(x),
                                   x.id == self.account.id)
                        for x in get_accounts_with_unapplied()])
        return options


def get_csv_rows(line_items: list[JournalEntryLineItem]) -> list[CSVRow]:
    """Composes and returns the CSV rows from the line items.

    :param line_items: The line items.
    :return: The CSV rows.
    """
    rows: list[CSVRow] = [CSVRow(gettext("Date"), gettext("Currency"),
                                 gettext("Description"), gettext("Amount"),
                                 gettext("Net Balance"))]
    rows.extend([CSVRow(x.journal_entry.date, x.currency.code,
                        x.description, x.amount, x.net_balance)
                 for x in line_items])
    return rows


class UnappliedOriginalLineItems(BaseReport):
    """The unapplied original line items."""

    def __init__(self, account: Account):
        """Constructs the unapplied original line items.

        :param account: The account.
        """
        self.__account: Account = account
        """The account."""
        offset_matcher: OffsetMatcher = OffsetMatcher(self.__account)
        self.__line_items: list[JournalEntryLineItem] \
            = offset_matcher.unapplied
        """The line items."""
        self.__is_mark_matches: bool \
            = can_edit() and len(offset_matcher.unmatched_offsets) > 0
        """Whether to mark the matched offsets."""

    def csv(self) -> Response:
        """Returns the report as CSV for download.

        :return: The response of the report for download.
        """
        filename: str = f"unapplied-{self.__account.code}.csv"
        return csv_download(filename, get_csv_rows(self.__line_items))

    def html(self) -> str:
        """Composes and returns the report as HTML.

        :return: The report as HTML.
        """
        pagination: Pagination[JournalEntryLineItem] \
            = Pagination[JournalEntryLineItem](self.__line_items,
                                               is_reversed=True)
        params: PageParams = PageParams(account=self.__account,
                                        is_mark_matches=self.__is_mark_matches,
                                        pagination=pagination,
                                        line_items=pagination.list)
        return render_template("accounting/report/unapplied.html",
                               report=params)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/7

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The income and expenses log.

"""
from datetime import date
from decimal import Decimal

import sqlalchemy as sa
from flask import url_for, render_template, Response
from sqlalchemy.orm import selectinload

from accounting import db
from accounting.locale import gettext
from accounting.models import Currency, Account, JournalEntry, \
    JournalEntryLineItem
from accounting.report.period import Period, PeriodChooser
from accounting.report.utils.base_page_params import BasePageParams
from accounting.report.utils.base_report import BaseReport
from accounting.report.utils.csv_export import BaseCSVRow, csv_download, \
    period_spec
from accounting.report.utils.option_link import OptionLink
from accounting.report.utils.report_chooser import ReportChooser
from accounting.report.utils.report_type import ReportType
from accounting.report.utils.urls import income_expenses_url
from accounting.utils.cast import be
from accounting.utils.current_account import CurrentAccount
from accounting.utils.pagination import Pagination


class ReportLineItem:
    """A line item in the report."""

    def __init__(self, line_item: JournalEntryLineItem | None = None):
        """Constructs the line item in the report.

        :param line_item: The journal entry line item.
        """
        self.is_brought_forward: bool = False
        """Whether this is the brought-forward line item."""
        self.is_total: bool = False
        """Whether this is the total line item."""
        self.date: date | None = None
        """The date."""
        self.account: Account | None = None
        """The account."""
        self.description: str | None = None
        """The description."""
        self.income: Decimal | None = None
        """The income amount."""
        self.expense: Decimal | None = None
        """The expense amount."""
        self.balance: Decimal | None = None
        """The balance."""
        self.note: str | None = None
        """The note."""
        self.url: str | None = None
        """The URL to the journal entry line item."""
        if line_item is not None:
            self.date = line_item.journal_entry.date
            self.account = line_item.account
            self.description = line_item.description
            self.income = None if line_item.is_debit else line_item.amount
            self.expense = line_item.amount if line_item.is_debit else None
            self.note = line_item.journal_entry.note
            self.url = url_for("accounting.journal-entry.detail",
                               journal_entry=line_item.journal_entry)


class LineItemCollector:
    """The line item collector."""

    def __init__(self, currency: Currency, account: CurrentAccount,
                 period: Period):
        """Constructs the line item collector.

        :param currency: The currency.
        :param account: The account.
        :param period: The period.
        """
        self.__currency: Currency = currency
        """The currency."""
        self.__account: CurrentAccount = account
        """The account."""
        self.__period: Period = period
        """The period"""
        self.brought_forward: ReportLineItem | None
        """The brought-forward line item."""
        self.line_items: list[ReportLineItem]
        """The line items."""
        self.total: ReportLineItem | None
        """The total line item."""
        self.brought_forward = self.__get_brought_forward()
        self.line_items = self.__query_line_items()
        self.total = self.__get_total()
        self.__populate_balance()

    def __get_brought_forward(self) -> ReportLineItem | None:
        """Queries, composes and returns the brought-forward line item.

        :return: The brought-forward line item, or None if the period starts
            from the beginning.
        """
        if self.__period.start is None:
            return None
        balance_func: sa.Function = sa.func.sum(sa.case(
            (JournalEntryLineItem.is_debit, JournalEntryLineItem.amount),
            else_=-JournalEntryLineItem.amount))
        select: sa.Select = sa.Select(balance_func)\
            .join(JournalEntry).join(Account)\
            .filter(be(JournalEntryLineItem.currency_code
                       == self.__currency.code),
                    self.__account_condition,
                    JournalEntry.date < self.__period.start)
        balance: int | None = db.session.scalar(select)
        if balance is None:
            return None
        line_item: ReportLineItem = ReportLineItem()
        line_item.is_brought_forward = True
        line_item.date = self.__period.start
        line_item.account = Account.accumulated_change()
        line_item.description = gettext("Brought forward")
        if balance > 0:
            line_item.income = balance
        elif balance < 0:
            line_item.expense = -balance
        line_item.balance = balance
        return line_item

    def __query_line_items(self) -> list[ReportLineItem]:
        """Queries and returns the line items.

        :return: The line items.
        """
        conditions: list[sa.BinaryExpression] \
            = [JournalEntryLineItem.currency_code == self.__currency.code,
               self.__account_condition]
        if self.__period.start is not None:
            conditions.append(JournalEntry.date >= self.__period.start)
        if self.__period.end is not None:
            conditions.append(JournalEntry.date <= self.__period.end)
        journal_entry_with_account: sa.Select = sa.Select(JournalEntry.id).\
            join(JournalEntryLineItem).join(Account).filter(*conditions)

        return [ReportLineItem(x)
                for x in JournalEntryLineItem.query
                .join(JournalEntry).join(Account)
                .filter(JournalEntryLineItem.journal_entry_id
                        .in_(journal_entry_with_account),
                        JournalEntryLineItem.currency_code
                        == self.__currency.code,
                        sa.not_(self.__account_condition))
                .order_by(JournalEntry.date,
                          JournalEntry.no,
                          JournalEntryLineItem.is_debit,
                          JournalEntryLineItem.no)
                .options(selectinload(JournalEntryLineItem.account),
                         selectinload(JournalEntryLineItem.journal_entry))]

    @property
    def __account_condition(self) -> sa.BinaryExpression:
        if self.__account.code == CurrentAccount.CURRENT_AL_CODE:
            return CurrentAccount.sql_condition()
        return Account.id == self.__account.id

    def __get_total(self) -> ReportLineItem | None:
        """Composes the total line item.

        :return: The total line item, or None if there is no data.
        """
        if self.brought_forward is None and len(self.line_items) == 0:
            return None
        line_item: ReportLineItem = ReportLineItem()
        line_item.is_total = True
        line_item.description = gettext("Total")
        line_item.income = sum([x.income for x in self.line_items
                                if x.income is not None])
        line_item.expense = sum([x.expense for x in self.line_items
                                 if x.expense is not None])
        line_item.balance = line_item.income - line_item.expense
        if self.brought_forward is not None:
            line_item.balance \
                = self.brought_forward.balance + line_item.balance
        return line_item

    def __populate_balance(self) -> None:
        """Populates the balance of the line items.

        :return: None.
        """
        balance: Decimal = 0 if self.brought_forward is None \
            else self.brought_forward.balance
        for line_item in self.line_items:
            if line_item.income is not None:
                balance = balance + line_item.income
            if line_item.expense is not None:
                balance = balance - line_item.expense
            line_item.balance = balance


class CSVRow(BaseCSVRow):
    """A row in the CSV."""

    def __init__(self, journal_entry_date: date | str | None,
                 account: str | None,
                 description: str | None,
                 income: str | Decimal | None,
                 expense: str | Decimal | None,
                 balance: str | Decimal | None,
                 note: str | None):
        """Constructs a row in the CSV.

        :param journal_entry_date: The journal entry date.
        :param account: The account.
        :param description: The description.
        :param income: The income.
        :param expense: The expense.
        :param balance: The balance.
        :param note: The note.
        """
        self.date: date | str | None = journal_entry_date
        """The date."""
        self.account: str | None = account
        """The account."""
        self.description: str | None = description
        """The description."""
        self.income: str | Decimal | None = income
        """The income."""
        self.expense: str | Decimal | None = expense
        """The expense."""
        self.balance: str | Decimal | None = balance
        """The balance."""
        self.note: str | None = note
        """The note."""

    @property
    def values(self) -> list[str | Decimal | None]:
        """Returns the values of the row.

        :return: The values of the row.
        """
        return [self.date, self.account, self.description,
                self.income, self.expense, self.balance, self.note]


class PageParams(BasePageParams):
    """The HTML page parameters."""

    def __init__(self, currency: Currency,
                 account: CurrentAccount,
                 period: Period,
                 has_data: bool,
                 pagination: Pagination[ReportLineItem],
                 brought_forward: ReportLineItem | None,
                 line_items: list[ReportLineItem],
                 total: ReportLineItem | None):
        """Constructs the HTML page parameters.

        :param currency: The currency.
        :param account: The account.
        :param period: The period.
        :param has_data: True if there is any data, or False otherwise.
        :param brought_forward: The brought-forward line item.
        :param line_items: The line items.
        :param total: The total line item.
        """
        self.currency: Currency = currency
        """The currency."""
        self.account: CurrentAccount = account
        """The account."""
        self.period: Period = period
        """The period."""
        self.__has_data: bool = has_data
        """True if there is any data, or False otherwise."""
        self.pagination: Pagination[ReportLineItem] = pagination
        """The pagination."""
        self.brought_forward: ReportLineItem | None = brought_forward
        """The brought-forward line item."""
        self.line_items: list[ReportLineItem] = line_items
        """The line items."""
        self.total: ReportLineItem | None = total
        """The total line item."""
        self.period_chooser: PeriodChooser = PeriodChooser(
            lambda x: income_expenses_url(currency, account, x))
        """The period chooser."""

    @property
    def has_data(self) -> bool:
        """Returns whether there is any data on the page.

        :return: True if there is any data, or False otherwise.
        """
        return self.__has_data

    @property
    def report_chooser(self) -> ReportChooser:
        """Returns the report chooser.

        :return: The report chooser.
        """
        if self.account.account is None:
            return ReportChooser(ReportType.INCOME_EXPENSES,
                                 currency=self.currency,
                                 account=Account.cash(),
                                 period=self.period)
        return ReportChooser(ReportType.INCOME_EXPENSES,
                             currency=self.currency,
                             account=self.account.account,
                             period=self.period)

    @property
    def currency_options(self) -> list[OptionLink]:
        """Returns the currency options.

        :return: The currency options.
        """
        return self._get_currency_options(
            lambda x: income_expenses_url(x, self.account, self.period),
            self.currency)

    @property
    def account_options(self) -> list[OptionLink]:
        """Returns the account options.

        :return: The account options.
        """
        current_al: CurrentAccount \
            = CurrentAccount.current_assets_and_liabilities()
        options: list[OptionLink] \
            = [OptionLink(str(current_al),
                          income_expenses_url(self.currency, current_al,
                                              self.period),
                          self.account.id == 0)]
        in_use: sa.Select = sa.Select(JournalEntryLineItem.account_id)\
            .join(Account)\
            .filter(be(JournalEntryLineItem.currency_code
                       == self.currency.code),
                    CurrentAccount.sql_condition())\
            .group_by(JournalEntryLineItem.account_id)
        options.extend([OptionLink(str(x),
                                   income_expenses_url(
                                       self.currency,
                                       CurrentAccount(x),
                                       self.period),
                                   x.id == self.account.id)
                        for x in Account.query.filter(Account.id.in_(in_use))
                       .order_by(Account.base_code, Account.no).all()])
        return options


class IncomeExpenses(BaseReport):
    """The income and expenses log."""

    def __init__(self, currency: Currency, account: CurrentAccount,
                 period: Period):
        """Constructs an income and expenses log.

        :param currency: The currency.
        :param account: The account.
        :param period: The period.
        """
        self.__currency: Currency = currency
        """The currency."""
        self.__account: CurrentAccount = account
        """The account."""
        self.__period: Period = period
        """The period."""
        collector: LineItemCollector = LineItemCollector(
            self.__currency, self.__account, self.__period)
        self.__brought_forward: ReportLineItem | None \
            = collector.brought_forward
        """The brought-forward line item."""
        self.__line_items: list[ReportLineItem] = collector.line_items
        """The line items."""
        self.__total: ReportLineItem | None = collector.total
        """The total line item."""

    def csv(self) -> Response:
        """Returns the report as CSV for download.

        :return: The response of the report for download.
        """
        filename: str = "income-expenses-{currency}-{account}-{period}.csv"\
            .format(currency=self.__currency.code, account=self.__account.code,
                    period=period_spec(self.__period))
        return csv_download(filename, self.__get_csv_rows())

    def __get_csv_rows(self) -> list[CSVRow]:
        """Composes and returns the CSV rows.

        :return: The CSV rows.
        """
        rows: list[CSVRow] = [CSVRow(gettext("Date"), gettext("Account"),
                                     gettext("Description"), gettext("Income"),
                                     gettext("Expense"), gettext("Balance"),
                                     gettext("Note"))]
        if self.__brought_forward is not None:
            rows.append(CSVRow(self.__brought_forward.date,
                               str(self.__brought_forward.account).title(),
                               self.__brought_forward.description,
                               self.__brought_forward.income,
                               self.__brought_forward.expense,
                               self.__brought_forward.balance,
                               None))
        rows.extend([CSVRow(x.date, str(x.account).title(), x.description,
                            x.income, x.expense, x.balance, x.note)
                     for x in self.__line_items])
        if self.__total is not None:
            rows.append(CSVRow(gettext("Total"), None, None,
                               self.__total.income, self.__total.expense,
                               self.__total.balance, None))
        return rows

    def html(self) -> str:
        """Composes and returns the report as HTML.

        :return: The report as HTML.
        """
        all_line_items: list[ReportLineItem] = []
        if self.__brought_forward is not None:
            all_line_items.append(self.__brought_forward)
        all_line_items.extend(self.__line_items)
        if self.__total is not None:
            all_line_items.append(self.__total)
        pagination: Pagination[ReportLineItem] \
            = Pagination[ReportLineItem](all_line_items, is_reversed=True)
        page_line_items: list[ReportLineItem] = pagination.list
        has_data: bool = len(page_line_items) > 0
        brought_forward: ReportLineItem | None = None
        if len(page_line_items) > 0 and page_line_items[0].is_brought_forward:
            brought_forward = page_line_items[0]
            page_line_items = page_line_items[1:]
        total: ReportLineItem | None = None
        if len(page_line_items) > 0 and page_line_items[-1].is_total:
            total = page_line_items[-1]
            page_line_items = page_line_items[:-1]
        params: PageParams = PageParams(currency=self.__currency,
                                        account=self.__account,
                                        period=self.__period,
                                        has_data=has_data,
                                        pagination=pagination,
                                        brought_forward=brought_forward,
                                        line_items=page_line_items,
                                        total=total)
        return render_template("accounting/report/income-expenses.html",
                               report=params)
